import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy
import scipy.optimize as optimize
from scipy.special import erf





## Anhydrous Molar masses, used for ONeill (2021). Doesnt include P2O5, Cr2O3 or H2O in the calculation.
oxide_mass_liq_anhyd = {'SiO2_Liq': 60.0843, 'MgO_Liq': 40.3044,
'MnO_Liq': 70.9375, 'FeOt_Liq': 71.8464, 'CaO_Liq': 56.0774,
'Al2O3_Liq': 101.961, 'Na2O_Liq': 61.9789, 'K2O_Liq': 94.196,
'TiO2_Liq': 79.8788, 'P2O5_Liq': 141.944522 }


cation_num_liq_anhyd = {'SiO2_Liq': 1, 'MgO_Liq': 1, 'MnO_Liq': 1,
'FeOt_Liq': 1, 'CaO_Liq': 1, 'Al2O3_Liq': 2, 'Na2O_Liq': 2,
'K2O_Liq': 2, 'TiO2_Liq': 1, 'P2O5_Liq': 2}


# Turns dictionary into a dataframe so pandas matrix math functions can be used
oxide_mass_liq_anhyd_df = pd.DataFrame.from_dict(
    oxide_mass_liq_anhyd, orient='index').T
oxide_mass_liq_anhyd_df['Sample_ID_Liq'] = 'MolWt'
oxide_mass_liq_anhyd_df.set_index('Sample_ID_Liq', inplace=True)



#
cation_num_liq_anhyd_df = pd.DataFrame.from_dict(
    cation_num_liq_anhyd, orient='index').T
cation_num_liq_anhyd_df['Sample_ID_Liq'] = 'CatNum'
cation_num_liq_anhyd_df.set_index('Sample_ID_Liq', inplace=True)

## Hydrous Molar masses - Used for Fortin et al. (2015)

oxide_mass_liq_hyd = {'SiO2_Liq': 60.0843, 'MgO_Liq': 40.3044,
'MnO_Liq': 70.9375, 'FeOt_Liq': 71.8464, 'CaO_Liq': 56.0774,
'Al2O3_Liq': 101.961, 'Na2O_Liq': 61.9789, 'K2O_Liq': 94.196,
'TiO2_Liq': 79.8788, 'P2O5_Liq': 141.944, 'H2O_Liq': 18}


cation_num_liq_hyd = {'SiO2_Liq': 1, 'MgO_Liq': 1, 'MnO_Liq': 1,
'FeOt_Liq': 1, 'CaO_Liq': 1, 'Al2O3_Liq': 2, 'Na2O_Liq': 2,
'K2O_Liq': 2, 'TiO2_Liq': 1, 'P2O5_Liq': 2, 'H2O_Liq':2}


# Turns dictionary into a dataframe so pandas matrix math functions can be used
oxide_mass_liq_hyd_df = pd.DataFrame.from_dict(
    oxide_mass_liq_hyd, orient='index').T
oxide_mass_liq_hyd_df['Sample_ID_Liq'] = 'MolWt'
oxide_mass_liq_hyd_df.set_index('Sample_ID_Liq', inplace=True)



#
cation_num_liq_hyd_df = pd.DataFrame.from_dict(
    cation_num_liq_hyd, orient='index').T
cation_num_liq_hyd_df['Sample_ID_Liq'] = 'CatNum'
cation_num_liq_hyd_df.set_index('Sample_ID_Liq', inplace=True)





## Anhydrous proportions, fractions, calculations from Thermobar (Wieser et al. in prep)


def calculate_anhydrous_mol_proportions_liquid(liq_comps):
    '''Import Liq compositions using liq_comps=My_Liquids, returns anhydrous mole proportions

   Parameters
    -------

    liq_comps: pandas.DataFrame
        Panda DataFrame of liquid compositions with column headings SiO2_Liq, TiO2_Liq etc.

    Returns
    -------
    pandas DataFrame
        anhydrous mole proportions for the liquid with column headings of the form SiO2_Liq_mol_prop

    '''
    # This makes the input match the columns in the oxide mass dataframe
    liq_wt = liq_comps.reindex(
        oxide_mass_liq_anhyd_df.columns, axis=1).fillna(0)
    # Combine the molecular weight and weight percent dataframes
    liq_wt_combo = pd.concat([oxide_mass_liq_anhyd_df, liq_wt],)
    # Drop the calculation column
    mol_prop_anhyd = liq_wt_combo.div(
        liq_wt_combo.loc['MolWt', :], axis='columns').drop(['MolWt'])
    mol_prop_anhyd.columns = [
        str(col) + '_mol_prop' for col in mol_prop_anhyd.columns]
    return mol_prop_anhyd


def calculate_anhydrous_mol_fractions_liquid(liq_comps):
    '''Import Liq compositions using liq_comps=My_Liquids, returns anhydrous mole fractions

   Parameters
    -------

    liq_comps: pandas.DataFrame
                Panda DataFrame of liquid compositions with column headings SiO2_Liq, TiO2_Liq etc.

    Returns
    -------
    pandas DataFrame
        anhydrous mole fractions for the liquid with column headings of the form SiO2_Liq_mol_frac

    '''
    mol_prop = calculate_anhydrous_mol_proportions_liquid(liq_comps)
    mol_prop['sum'] = mol_prop.sum(axis='columns')
    mol_frac_anhyd = mol_prop.div(mol_prop['sum'], axis='rows')
    mol_frac_anhyd.drop(['sum'], axis='columns', inplace=True)
    mol_frac_anhyd.columns = [str(col).replace('prop', 'frac')
                              for col in mol_frac_anhyd.columns]
    return mol_frac_anhyd


def calculate_anhydrous_cat_proportions_liquid(liq_comps):
    '''Import Liq compositions using liq_comps=My_Liquids, returns anhydrous cation proportions (e.g., mole proportions * no of cations)

   Parameters
    -------

    liq_comps: pandas.DataFrame
                Panda DataFrame of liquid compositions with column headings SiO2_Liq, TiO2_Liq etc.


    Returns
    -------
    pandas DataFrame
        anhydrous cation proportions for the liquid with column headings of the form S_Liq_cat_prop

    '''
    mol_prop_no_cat_num = calculate_anhydrous_mol_proportions_liquid(liq_comps)
    mol_prop_no_cat_num.columns = [str(col).replace(
        '_mol_prop', '') for col in mol_prop_no_cat_num.columns]
    ox_num_reindex = cation_num_liq_anhyd_df.reindex(
        oxide_mass_liq_anhyd_df.columns, axis=1).fillna(0)
    df_calc_comb = pd.concat([ox_num_reindex, mol_prop_no_cat_num])
    cation_prop_anhyd = df_calc_comb.multiply(
        df_calc_comb.loc['CatNum', :], axis='columns').drop(['CatNum'])
    cation_prop_anhyd.columns = [
        str(col) + '_cat_prop' for col in cation_prop_anhyd.columns]



    return cation_prop_anhyd



def calculate_anhydrous_cat_fractions_liquid(liq_comps):
    '''Import Liq compositions using liq_comps=My_Liquids, returns anhydrous cation fractions

   Parameters
    -------

    liq_comps: pandas.DataFrame
        liquid compositions with column headings SiO2_Liq, TiO2_Liq etc.


    Returns
    -------
    pandas DataFrame
        anhydrous cation fractions for the liquid with column headings of the form _Liq_cat_frac,
        as well as the initial dataframe of liquid compositions.


    '''
    cat_prop = calculate_anhydrous_cat_proportions_liquid(liq_comps=liq_comps)
    mol_prop = calculate_anhydrous_mol_fractions_liquid(liq_comps=liq_comps)
    cat_prop['sum'] = cat_prop.sum(axis='columns')
    cat_frac_anhyd = cat_prop.div(cat_prop['sum'], axis='rows')
    cat_frac_anhyd.drop(['sum'], axis='columns', inplace=True)
    cat_frac_anhyd.columns = [str(col).replace('prop', 'frac')
                              for col in cat_frac_anhyd.columns]
    cat_frac_anhyd = pd.concat([liq_comps, mol_prop, cat_frac_anhyd], axis=1)

    if "Fe3Fet_Liq" in cat_frac_anhyd:
        cat_frac_anhyd['Mg_Number_Liq_NoFe3'] = (cat_frac_anhyd['MgO_Liq'] / 40.3044) / (
            (cat_frac_anhyd['MgO_Liq'] / 40.3044) + (cat_frac_anhyd['FeOt_Liq'] / 71.844))
        cat_frac_anhyd['Mg_Number_Liq_Fe3'] = (cat_frac_anhyd['MgO_Liq'] / 40.3044) / (
            (cat_frac_anhyd['MgO_Liq'] / 40.3044) + (cat_frac_anhyd['FeOt_Liq'] * (1 - cat_frac_anhyd['Fe3Fet_Liq']) / 71.844))
    if "Fe3Fet_Liq" not in cat_frac_anhyd:
        cat_frac_anhyd['Mg_Number_Liq_Fe3'] = (cat_frac_anhyd['MgO_Liq'] / 40.3044) / (
            (cat_frac_anhyd['MgO_Liq'] / 40.3044) + (cat_frac_anhyd['FeOt_Liq'] / 71.844))
        cat_frac_anhyd['Mg_Number_Liq_NoFe3'] = (cat_frac_anhyd['MgO_Liq'] / 40.3044) / (
            (cat_frac_anhyd['MgO_Liq'] / 40.3044) + (cat_frac_anhyd['FeOt_Liq'] / 71.844))


    cation_frac_anhyd2=cat_frac_anhyd.rename(columns={
                        'SiO2_Liq_cat_frac': 'Si_Liq_cat_frac',
                        'TiO2_Liq_cat_frac': 'Ti_Liq_cat_frac',
                        'Al2O3_Liq_cat_frac': 'Al_Liq_cat_frac',
                        'FeOt_Liq_cat_frac': 'Fet_Liq_cat_frac',
                        'MnO_Liq_cat_frac': 'Mn_Liq_cat_frac',
                        'MgO_Liq_cat_frac': 'Mg_Liq_cat_frac',
                        'CaO_Liq_cat_frac': 'Ca_Liq_cat_frac',
                        'Na2O_Liq_cat_frac': 'Na_Liq_cat_frac',
                        'K2O_Liq_cat_frac': 'K_Liq_cat_frac',
                        'Cr2O3_Liq_cat_frac': 'Cr_Liq_cat_frac',
                        'P2O5_Liq_cat_frac': 'P_Liq_cat_frac',

                        })

    return cation_frac_anhyd2


## hydrous proportions, fractions, calculations from Thermobar (Wieser et al. in prep)


def calculate_hydrous_mol_proportions_liquid(liq_comps):
    '''Import Liq compositions using liq_comps=My_Liquids, returns hydrous mole proportions

   Parameters
    -------

    liq_comps: pandas.DataFrame
        Panda DataFrame of liquid compositions with column headings SiO2_Liq, TiO2_Liq etc.

    Returns
    -------
    pandas DataFrame
        hydrous mole proportions for the liquid with column headings of the form SiO2_Liq_mol_prop

    '''
    # This makes the input match the columns in the oxide mass dataframe
    liq_wt = liq_comps.reindex(
        oxide_mass_liq_hyd_df.columns, axis=1).fillna(0)
    # Combine the molecular weight and weight percent dataframes
    liq_wt_combo = pd.concat([oxide_mass_liq_hyd_df, liq_wt],)
    # Drop the calculation column
    mol_prop_hyd = liq_wt_combo.div(
        liq_wt_combo.loc['MolWt', :], axis='columns').drop(['MolWt'])
    mol_prop_hyd.columns = [
        str(col) + '_mol_prop' for col in mol_prop_hyd.columns]
    return mol_prop_hyd


def calculate_hydrous_mol_fractions_liquid(liq_comps):
    '''Import Liq compositions using liq_comps=My_Liquids, returns hydrous mole fractions

   Parameters
    -------

    liq_comps: pandas.DataFrame
                Panda DataFrame of liquid compositions with column headings SiO2_Liq, TiO2_Liq etc.

    Returns
    -------
    pandas DataFrame
        hydrous mole fractions for the liquid with column headings of the form SiO2_Liq_mol_frac

    '''
    mol_prop = calculate_hydrous_mol_proportions_liquid(liq_comps)
    mol_prop['sum'] = mol_prop.sum(axis='columns')
    mol_frac = mol_prop.div(mol_prop['sum'], axis='rows')
    mol_frac.drop(['sum'], axis='columns', inplace=True)
    mol_frac.columns = [str(col).replace('prop', 'frac')
                              for col in mol_frac.columns]
    return mol_frac


def calculate_hydrous_cat_proportions_liquid(liq_comps):
    '''Import Liq compositions using liq_comps=My_Liquids, returns hydrous cation proportions (e.g., mole proportions * no of cations)

   Parameters
    -------

    liq_comps: pandas.DataFrame
                Panda DataFrame of liquid compositions with column headings SiO2_Liq, TiO2_Liq etc.


    Returns
    -------
    pandas DataFrame
        hydrous cation proportions for the liquid with column headings of the form S_Liq_cat_prop

    '''
    mol_prop_no_cat_num = calculate_hydrous_mol_proportions_liquid(liq_comps)
    mol_prop_no_cat_num.columns = [str(col).replace(
        '_mol_prop', '') for col in mol_prop_no_cat_num.columns]
    ox_num_reindex = cation_num_liq_hyd_df.reindex(
        oxide_mass_liq_hyd_df.columns, axis=1).fillna(0)
    df_calc_comb = pd.concat([ox_num_reindex, mol_prop_no_cat_num])
    cation_prop_hyd = df_calc_comb.multiply(
        df_calc_comb.loc['CatNum', :], axis='columns').drop(['CatNum'])
    cation_prop_hyd.columns = [
        str(col) + '_cat_prop' for col in cation_prop_hyd.columns]



    return cation_prop_hyd



def calculate_hydrous_cat_fractions_liquid(liq_comps):
    '''Import Liq compositions using liq_comps=My_Liquids, returns hydrous cation fractions

   Parameters
    -------

    liq_comps: pandas.DataFrame
        liquid compositions with column headings SiO2_Liq, TiO2_Liq etc.


    Returns
    -------
    pandas DataFrame
        hydrous cation fractions for the liquid with column headings of the form _Liq_cat_frac,
        as well as the initial dataframe of liquid compositions.


    '''
    cat_prop = calculate_hydrous_cat_proportions_liquid(liq_comps=liq_comps)
    mol_prop = calculate_hydrous_mol_fractions_liquid(liq_comps=liq_comps)
    cat_prop['sum'] = cat_prop.sum(axis='columns')
    cat_frac_hyd = cat_prop.div(cat_prop['sum'], axis='rows')
    cat_frac_hyd.drop(['sum'], axis='columns', inplace=True)
    cat_frac_hyd.columns = [str(col).replace('prop', 'frac')
                              for col in cat_frac_hyd.columns]
    cat_frac_hyd = pd.concat([liq_comps, mol_prop, cat_frac_hyd], axis=1)

    if "Fe3Fet_Liq" in cat_frac_hyd:
        cat_frac_hyd['Mg_Number_Liq_NoFe3'] = (cat_frac_hyd['MgO_Liq'] / 40.3044) / (
            (cat_frac_hyd['MgO_Liq'] / 40.3044) + (cat_frac_hyd['FeOt_Liq'] / 71.844))
        cat_frac_hyd['Mg_Number_Liq_Fe3'] = (cat_frac_hyd['MgO_Liq'] / 40.3044) / (
            (cat_frac_hyd['MgO_Liq'] / 40.3044) + (cat_frac_hyd['FeOt_Liq'] * (1 - cat_frac_hyd['Fe3Fet_Liq']) / 71.844))
    if "Fe3Fet_Liq" not in cat_frac_hyd:
        cat_frac_hyd['Mg_Number_Liq_Fe3'] = (cat_frac_hyd['MgO_Liq'] / 40.3044) / (
            (cat_frac_hyd['MgO_Liq'] / 40.3044) + (cat_frac_hyd['FeOt_Liq'] / 71.844))
        cat_frac_hyd['Mg_Number_Liq_NoFe3'] = (cat_frac_hyd['MgO_Liq'] / 40.3044) / (
            (cat_frac_hyd['MgO_Liq'] / 40.3044) + (cat_frac_hyd['FeOt_Liq'] / 71.844))


    cation_frac_hyd2=cat_frac_hyd.rename(columns={
                        'SiO2_Liq_cat_frac': 'Si_Liq_cat_frac',
                        'TiO2_Liq_cat_frac': 'Ti_Liq_cat_frac',
                        'Al2O3_Liq_cat_frac': 'Al_Liq_cat_frac',
                        'FeOt_Liq_cat_frac': 'Fet_Liq_cat_frac',
                        'MnO_Liq_cat_frac': 'Mn_Liq_cat_frac',
                        'MgO_Liq_cat_frac': 'Mg_Liq_cat_frac',
                        'CaO_Liq_cat_frac': 'Ca_Liq_cat_frac',
                        'Na2O_Liq_cat_frac': 'Na_Liq_cat_frac',
                        'K2O_Liq_cat_frac': 'K_Liq_cat_frac',
                        'Cr2O3_Liq_cat_frac': 'Cr_Liq_cat_frac',
                        'P2O5_Liq_cat_frac': 'P_Liq_cat_frac',

                        })

    return cation_frac_hyd2

## Hydrous Calculations - Used for Smythe et al. (2017)
oxide_mass_liq = {'SiO2_Liq': 60.0843, 'MgO_Liq': 40.3044,
'MnO_Liq': 70.9375, 'FeO_Liq': 71.8464, 'Fe2O3_Liq': 159.69, 'CaO_Liq': 56.0774,
'Al2O3_Liq': 101.961, 'Na2O_Liq': 61.9789, 'K2O_Liq': 94.196,
'TiO2_Liq': 79.8788, 'P2O5_Liq': 141.944}

oxide_mass_liq_df = pd.DataFrame.from_dict(
    oxide_mass_liq, orient='index').T
oxide_mass_liq_df['Sample_ID_Liq'] = 'MolWt'
oxide_mass_liq_df.set_index('Sample_ID_Liq', inplace=True)




def calculate_mol_proportions_liquid(liq_comps):
    '''Import Liq compositions using liq_comps=My_Liquids, returns anhydrous mole proportions

   Parameters
    -------

    liq_comps: pandas.DataFrame
        Panda DataFrame of liquid compositions with column headings SiO2_Liq, TiO2_Liq etc.

    Returns
    -------
    pandas DataFrame
        anhydrous mole proportions for the liquid with column headings of the form SiO2_Liq_mol_prop

    '''
    # This makes the input match the columns in the oxide mass dataframe
    liq_wt = liq_comps.reindex(
        oxide_mass_liq_df.columns, axis=1).fillna(0)
    # Combine the molecular weight and weight percent dataframes
    liq_wt_combo = pd.concat([oxide_mass_liq_df, liq_wt],)
    # Drop the calculation column
    mol_prop_anhyd = liq_wt_combo.div(
        liq_wt_combo.loc['MolWt', :], axis='columns').drop(['MolWt'])
    mol_prop_anhyd.columns = [
        str(col) + '_mol_prop' for col in mol_prop_anhyd.columns]
    return mol_prop_anhyd




# For oxide to wt% function
oxide_mass_all = {'SiO2_Liq': 60.084, 'MgO_Liq': 40.304, 'FeO_Liq': 71.846, 'Fe2O3_Liq':159.69,
    'CaO_Liq': 56.079, 'Al2O3_Liq': 101.961, 'Na2O_Liq': 61.979, 'K2O_Liq': 94.195, 'MnO_Liq': 70.937, 'TiO2_Liq': 79.898,
                          'P2O5_Liq':141.944, 'H2O_Liq': 18.01528}
oxide_mass_all = pd.DataFrame.from_dict(
    oxide_mass_all, orient='index').T
oxide_mass_all['Sample_ID'] = 'MolWt'
oxide_mass_all.set_index('Sample_ID', inplace=True)

elemental_mass_mult_all = {'SiO2_Liq': 28.0855, 'MgO_Liq': 24.305, 'FeO_Liq': 55.845,
'Fe2O3_Liq': 55.845*2, 'CaO_Liq': 40.078, 'Al2O3_Liq': 26.981539*2,
                          'Na2O_Liq': 22.989769*2, 'K2O_Liq': 39.0983*2, 'MnO_Liq': 54.938044, 'TiO2_Liq': 47.867,
                       'P2O5_Liq': 2*30.973762, 'H2O_Liq':1.00794*2}
elemental_mass_mult_all = pd.DataFrame.from_dict(
    elemental_mass_mult_all, orient='index').T
elemental_mass_mult_all['Sample_ID'] = 'ElWt'
elemental_mass_mult_all.set_index('Sample_ID', inplace=True)



df_ideal_all2 = pd.DataFrame(columns=['SiO2', 'TiO2', 'Al2O3',
'FeO', 'Fe2O3', 'MnO', 'MgO', 'CaO', 'Na2O', 'K2O',
'Cr2O3', 'P2O5', 'F', 'Cl', 'H2O'])

def convert_oxide_percent_to_element_weight_percent(df, Fe3Fet_Liq=None):


    df_c=df.copy()
    if Fe3Fet_Liq is not None:
        df_c['Fe3Fet_Liq']=Fe3Fet_Liq

    df_c['FeO_Liq']=df_c['FeOt_Liq']*(1-df_c['Fe3Fet_Liq'])
    df_c['Fe2O3_Liq']=df_c['FeOt_Liq']*(df_c['Fe3Fet_Liq'])*1.11111
    df_c.drop(columns=['FeOt_Liq'], inplace=True)

    df_oxides=df_c.reindex(oxide_mass_all.columns, axis=1).fillna(0)

    liq_wt_combo = pd.concat([oxide_mass_all, df_oxides],)


    mol_prop_anhyd = liq_wt_combo.div(
        liq_wt_combo.loc['MolWt', :], axis='columns').drop(['MolWt'])

    el_combo=pd.concat([elemental_mass_mult_all, mol_prop_anhyd ],)
    wt_perc = el_combo.multiply(
        el_combo.loc['ElWt', :], axis='columns').drop(['ElWt'])


    wt_perc2=pd.DataFrame(data={'Si_wt': wt_perc['SiO2_Liq'],
                                'Mg_wt': wt_perc['MgO_Liq'],
                                'Fe2_wt':wt_perc['FeO_Liq'],
                                'Fe3_wt':wt_perc['Fe2O3_Liq'],
                                'Ca_wt':wt_perc['CaO_Liq'],
                                'Al_wt':wt_perc['Al2O3_Liq'],
                                'Na_wt':wt_perc['Na2O_Liq'],
                                'K_wt':wt_perc['K2O_Liq'],
                                'Mn_wt':wt_perc['MnO_Liq'],
                                'Ti_wt':wt_perc['TiO2_Liq'],
                                'P_wt':wt_perc['P2O5_Liq'],
                                'H_wt': wt_perc['H2O_Liq'],



                                })
    sum_element=wt_perc2.sum(axis=1)

    mol_prop=calculate_mol_proportions_liquid(liq_comps=df_c)



    O_by_Charge=(15.9994/2*(mol_prop['SiO2_Liq_mol_prop']*4+mol_prop['TiO2_Liq_mol_prop']*4
    +2*mol_prop['Al2O3_Liq_mol_prop']*3+mol_prop['MgO_Liq_mol_prop']*2+mol_prop['MnO_Liq_mol_prop']*2+mol_prop['FeO_Liq_mol_prop']*2
    +2*mol_prop['Fe2O3_Liq_mol_prop']*3+mol_prop['CaO_Liq_mol_prop']*2+2*mol_prop['Na2O_Liq_mol_prop']*1+2*mol_prop['K2O_Liq_mol_prop']*1
    +2*mol_prop['P2O5_Liq_mol_prop']*5+df['H2O_Liq']/9.00764))

    wt_perc2['O_wt']=O_by_Charge

    return wt_perc2

atomic_wt=pd.DataFrame(data={'Sample_ID_Liq': "Atom_wt",
                                'Si_wt': 28.0855,
                                'Ti_wt': 47.867,
                                'Al_wt': 26.981538,
                                'Mg_wt':24.305,
                                'Mn_wt':54.938049,
                                'Fe2_wt':55.8457,
                                'Fe3_wt': 55.8457,
                                'Ca_wt': 40.078,
                                'Na_wt': 22.98977,
                                'K_wt': 39.0983,
                                'P_wt': 30.973761,
                                'H_wt': 1.00794,
                                'O_wt': 15.9994
                                    }, index=[0])

atomic_wt.set_index('Sample_ID_Liq', inplace=True)

def calculate_Smythe_silicate_mole_fractions(df, Fe3Fet_Liq=None):
    '''Smythe et al. call the results of these calculations "Silicate Mole Fractions"


   Parameters
    -------

    liq_comps: pandas.DataFrame
        Panda DataFrame of liquid compositions with column headings SiO2_Liq, TiO2_Liq etc.

    Returns
    -------
    pandas DataFrame
        anhydrous mole proportions for the liquid with column headings of the form SiO2_Liq_mol_prop

    '''
    if Fe3Fet_Liq is not None:
        el_wt=convert_oxide_percent_to_element_weight_percent(df, Fe3Fet_Liq)
    else:
        el_wt=convert_oxide_percent_to_element_weight_percent(df)


    # This makes the input match the columns in the oxide mass dataframe
    liq_wt = el_wt.reindex(
        atomic_wt.columns, axis=1).fillna(0)
    # Combine the molecular weight and weight percent dataframes
    liq_wt_combo = pd.concat([atomic_wt, liq_wt],)
    # Drop the calculation column
    atom_prop = liq_wt_combo.div(
        liq_wt_combo.loc['Atom_wt', :], axis='columns').drop(['Atom_wt'])
    atom_prop.columns = [
        str(col) + '_atom' for col in atom_prop.columns]

    atom_prop['sum'] = atom_prop.sum(axis='columns')
    atom_prop2 = atom_prop.div(atom_prop['sum'], axis='rows')
    atom_prop2.drop(['sum'], axis='columns', inplace=True)
    # atom_prop2.columns = [str(col).replace('Atom_wt', 'Atom_perc')
    #                           for col in Atom_wt.columns]
    a=atom_prop2*100

    a_drop=a.drop(columns=['O_wt_atom'])
    a_drop['sum'] = a_drop.sum(axis='columns')
    Final_Calc_step = a_drop.div(a_drop['sum'], axis='rows')
    Final_Calc_step.drop(['sum'], axis='columns', inplace=True)


    return Final_Calc_step

## Redox calculations
def convert_fo2_to_fe_partition(*, liq_comps, T_K, P_kbar,
                                model="Kress1991", fo2, renorm=False, fo2_offset=0):
    '''
    Calculates Fe3Fet_Liq, FeO and Fe2O3 based on user-specified buffer

   Parameters
    -------

    liq_comps: pandas.DataFrame
        Liquid compositions with column headings SiO2_Liq, MgO_Liq etc.

    T_K:  int, flt, pandas.Series
        Temperature in Kelvin (buffer positions are very T-sensitive)

    P_kbar: int, flt, pandas.Series
        Pressure in Kbar (Buffer positions are slightly sensitive to pressure)

    fo2:  str ("QFM", "NNO") or int, flt, pandas.Series
        Either a value of fo2 (enter 10*logfo2), or buffer position.
        So far, includes QFM or NNO

    fo2_offset: int, flt, pandas.Series
        log units offset from buffer, e.g., could specify fo2=QFM, fo2_offset=1
        to perform calculations at QFM+1

    model: str
        "Kress1991" - Uses Kress and Carmichael 1991 to calculate XFe2Fe3 from fo2
        "Put2016_eq6b" - Uses Putirka (2016) expression to calculate XFe2Fe3 from fo2

    renorm: bool
        Following excel code of K. Iacovino.
        If True, renormalizes other oxide concentrations
        to account for change in total following partitioning of Fe into FeO and Fe2O3.

    Returns
    -------

    liquid compositions with calculated Fe3Fet_Liq, FeO_Liq, Fe2O3_Liq, and XFe3Fe2.

    '''
    if isinstance(fo2, str):
        if fo2=="NNO":
        # Buffer position from frost (1991)
            logfo2=(-24930/T_K) + 9.36 + 0.046 * ((P_kbar*1000)-1)/T_K+fo2_offset
            fo2=10**logfo2

        if fo2=="QFM":
        # Buffer position from frost (1991)
            logfo2=(-25096.3/T_K) + 8.735 + 0.11 * ((P_kbar*1000)-1)/T_K+fo2_offset
            fo2=10**logfo2



    liq_comps_c=liq_comps.copy()
    mol_frac_short=calculate_hydrous_mol_fractions_liquid(liq_comps_c)
    mol_frac=pd.concat([mol_frac_short, liq_comps_c], axis=1)
    To=1673.15

    if model=="Kress1991":
        ln_XFe2FeO3_XFeO=((0.196*np.log(fo2))+(11492/T_K)-6.675+((-2.243*mol_frac['Al2O3_Liq_mol_frac'])+(-1.828*mol_frac['FeOt_Liq_mol_frac'])
        +(3.201*mol_frac['CaO_Liq_mol_frac'])+(5.854*mol_frac['Na2O_Liq_mol_frac'])+(6.215*mol_frac['K2O_Liq_mol_frac']))
        -3.36*(1-(To/T_K) - np.log(T_K/To)) -0.000000701*((P_kbar*100000000)/T_K)
         + -0.000000000154*(((T_K-1673)*(P_kbar*100000000))/T_K) + 0.0000000000000000385*((P_kbar*100000000)**2/T_K))
        print(ln_XFe2FeO3_XFeO)
    if model=="Put2016_eq6b":
        ln_XFe2FeO3_XFeO=(-6.35+10813.8/T_K + 0.19*np.log(fo2)+ 12.4*(mol_frac['Na2O_Liq_mol_frac']
         +mol_frac['K2O_Liq_mol_frac'])
        -3.44*(mol_frac['Al2O3_Liq_mol_frac']/(mol_frac['Al2O3_Liq_mol_frac']+mol_frac['SiO2_Liq_mol_frac']))
        +4.15*mol_frac['CaO_Liq_mol_frac'])

    X_Fe2O3_X_FeO=np.exp(ln_XFe2FeO3_XFeO)
    X_Fe2O3=X_Fe2O3_X_FeO*mol_frac['FeOt_Liq_mol_frac']/(2*X_Fe2O3_X_FeO+1)

    #X_FeO=mol_frac['FeOt_Liq_mol_frac']/(2*X_Fe2O3_X_FeO+1) Kayla's way
    X_FeO=mol_frac['FeOt_Liq_mol_frac']-2*X_Fe2O3
    Sum_all_mol_frac=(mol_frac['SiO2_Liq_mol_frac']+mol_frac['TiO2_Liq_mol_frac']+mol_frac['Al2O3_Liq_mol_frac']+mol_frac['MnO_Liq_mol_frac']
                      +mol_frac['MgO_Liq_mol_frac']+mol_frac['CaO_Liq_mol_frac']+mol_frac['Na2O_Liq_mol_frac']+mol_frac['K2O_Liq_mol_frac']
                      +mol_frac['P2O5_Liq_mol_frac']+X_FeO+X_Fe2O3)

    Fe2O3_unnorm=X_Fe2O3*159.6
    FeO_unnorm=X_FeO*71.844
    Sum_All_mol=(mol_frac['SiO2_Liq_mol_frac']*60.0843+mol_frac['TiO2_Liq_mol_frac']*79.8788
    +mol_frac['Al2O3_Liq_mol_frac']*101.961+mol_frac['MnO_Liq_mol_frac']*70.9375
    +mol_frac['MgO_Liq_mol_frac']*40.3044+mol_frac['CaO_Liq_mol_frac']*56.0774+mol_frac['Na2O_Liq_mol_frac']*61.9789+mol_frac['K2O_Liq_mol_frac']*94.196
    +mol_frac['P2O5_Liq_mol_frac']*141.937+X_Fe2O3*159.6+X_FeO*71.844)
    New_Fe2O3_wt=(100*X_Fe2O3*159.6)/Sum_All_mol
    New_FeO_wt=(100*X_FeO*71.844)/Sum_All_mol

    New_Oxide_out_nonorm=liq_comps.copy()
    New_Oxide_out_nonorm['FeO_Liq']=New_FeO_wt
    New_Oxide_out_nonorm['Fe2O3_Liq']=New_Fe2O3_wt
    New_Oxide_out_nonorm['XFe3Fe2']=X_Fe2O3_X_FeO
    New_Oxide_out_nonorm['Fe3Fet_Liq']=New_Fe2O3_wt*0.8998/(New_FeO_wt+New_Fe2O3_wt*0.8998)


    New_Oxide_out_norm=pd.DataFrame(data={'SiO2_Liq': 100*mol_frac['SiO2_Liq_mol_frac']*60.084/Sum_All_mol,
                                         'TiO2_Liq': 100*mol_frac['TiO2_Liq_mol_frac']*79.8788/Sum_All_mol,
                                         'Al2O3_Liq':100*mol_frac['Al2O3_Liq_mol_frac']*101.961/Sum_All_mol,
                                          'Fe2O3_Liq': (100*X_Fe2O3*159.6)/Sum_All_mol,
                                          'FeO_Liq': (100*X_FeO*71.844)/Sum_All_mol,
                                          'MnO_Liq': 100*mol_frac['MnO_Liq_mol_frac']*70.9375/Sum_All_mol,
                                          'MgO_Liq': 100*mol_frac['MgO_Liq_mol_frac']*40.3044/Sum_All_mol,
                                         'CaO_Liq': 100*mol_frac['CaO_Liq_mol_frac']*56.0774/Sum_All_mol,
                                          'Na2O_Liq': 100*mol_frac['Na2O_Liq_mol_frac']*61.9789/Sum_All_mol,
                                          'K2O_Liq': 100*mol_frac['K2O_Liq_mol_frac']*94.196/Sum_All_mol,
                                         'P2O5_Liq':  100*mol_frac['P2O5_Liq_mol_frac']*141.937/Sum_All_mol,
                                         })
    Old_Sum=(100/liq_comps_c.drop(['Sample_ID_Liq'], axis=1).sum(axis=1))
    New_Oxide_out_New_old_total=New_Oxide_out_norm.div(Old_Sum, axis=0)
    New_Oxide_out_New_old_total['Fe3Fet_Liq']=(New_Oxide_out_norm['Fe2O3_Liq']*0.8998/(New_Oxide_out_norm['FeO_Liq']+New_Oxide_out_norm['Fe2O3_Liq']*0.8998)).fillna(0)



    if renorm==False:
        return New_Oxide_out_nonorm
    else:
        return New_Oxide_out_New_old_total


## Need some functions for calculating mole proportions with Fe partition
oxide_mass_liq_hyd_redox = {'SiO2_Liq': 60.0843, 'MgO_Liq': 40.3044,
'MnO_Liq': 70.9375, 'FeO_Liq': 71.844, 'Fe2O3_Liq': 159.69, 'CaO_Liq': 56.0774,
'Al2O3_Liq': 101.961,'Na2O_Liq': 61.9789, 'K2O_Liq': 94.196,
 'TiO2_Liq': 79.8788, 'P2O5_Liq': 141.937, 'Cr2O3_Liq': 151.9982,
  'H2O_Liq': 18.01528}
# Turns dictionary into a dataframe so pandas matrix math functions can be used
oxide_mass_liq_hyd_df_redox = pd.DataFrame.from_dict(
    oxide_mass_liq_hyd_redox, orient='index').T
oxide_mass_liq_hyd_df_redox['Sample_ID_Liq'] = 'MolWt'
oxide_mass_liq_hyd_df_redox.set_index('Sample_ID_Liq', inplace=True)

def calculate_hydrous_mol_proportions_liquid_redox(liq_comps):
    '''Import Liq compositions using liq_comps=My_Liquids, returns anhydrous mole proportions

   Parameters
    -------


    liq_comps: pandas.DataFrame
        liquid compositions with column headings SiO2_Liq, TiO2_Liq etc.

    Returns
    -------
    pandas DataFrame
        anhydrous mole proportions for the liquid with column headings of the ..Liq_mol_prop

    '''
    # This makes the input match the columns in the oxide mass dataframe
    liq_wt = liq_comps.reindex(oxide_mass_liq_hyd_df_redox.columns, axis=1).fillna(0)
    # Combine the molecular weight and weight percent dataframes
    liq_wt_combo = pd.concat([oxide_mass_liq_hyd_df_redox, liq_wt],)
    # Drop the calculation column
    mol_prop_hyd = liq_wt_combo.div(
        liq_wt_combo.loc['MolWt', :], axis='columns').drop(['MolWt'])
    mol_prop_hyd.columns = [
        str(col) + '_mol_prop' for col in mol_prop_hyd.columns]
    return mol_prop_hyd

def calculate_hydrous_mol_fractions_liquid_redox(liq_comps):
    '''Import Liq compositions using liq_comps=My_Liquids, returns anhydrous mole fractions

   Parameters
    -------

    liq_comps: pandas.DataFrame
        liquid compositions with column headings SiO2_Liq, TiO2_Liq etc.



    Returns
    -------
    pandas DataFrame
        anhydrous mole fractions for the liquid with column headings of the form SiO2_Liq_mol_frac

    '''
    mol_prop = calculate_hydrous_mol_proportions_liquid_redox(liq_comps)
    mol_prop['sum'] = mol_prop.sum(axis='columns')
    mol_frac_hyd = mol_prop.div(mol_prop['sum'], axis='rows')
    mol_frac_hyd.drop(['sum'], axis='columns', inplace=True)
    mol_frac_hyd.columns = [str(col).replace('prop', 'frac')
                            for col in mol_frac_hyd.columns]
    return mol_frac_hyd



def convert_fe_partition_to_fo2(*, liq_comps, T_K, P_kbar,  model="Kress1991", renorm=False):
    '''
    Calculates delta fo2 relative to QFM and NNO buffer for liq compositions with FeO and Fe2O3

   Parameters
    -------

    liq_comps: pandas.DataFrame
        Liquid compositions with column headings SiO2_Liq, MgO_Liq, FeO_Liq and Fe2O3_Liq etc.

    T_K:  int, flt, pandas.Series
        Temperature in Kelvin (buffer positions are very T-sensitive)

    P_kbar: int, flt, pandas.Series
        Pressure in Kbar (Buffer positions are slightly sensitive to pressure)






    model: str
        "Kress1991" - Uses Kress and Carmichael 1991 to calculate XFe2Fe3 from fo2
        "Put2016_eq6b" - Uses Putirka (2016) expression to calculate XFe2Fe3 from fo2

    renorm: bool
        Following excel code of K. Iacovino.
        If True, renormalizes other oxide concentrations
        to account for change in total following partitioning of Fe into FeO and Fe2O3.

    Returns
    -------

    liquid compositions with calculated Fe3Fet_Liq, FeO_Liq, Fe2O3_Liq, and XFe3Fe2.

    '''

    liq_comps_c=liq_comps.copy()
    hyd_mol_frac_test=calculate_hydrous_mol_fractions_liquid(liq_comps=liq_comps_c)

    liq_comps_c['FeO_Liq']=liq_comps_c['FeOt_Liq']*(1-liq_comps_c['Fe3Fet_Liq'])
    liq_comps_c['Fe2O3_Liq']=liq_comps_c['FeOt_Liq']*(liq_comps_c['Fe3Fet_Liq'])*1.11111

    mol_frac_hyd_redox=calculate_hydrous_mol_fractions_liquid_redox(liq_comps=liq_comps_c)

    # Calculating buffer positions from Frost 1991

    To= 1673.15

    logfo2_NNO=(-24930/T_K) + 9.36 + 0.046 * ((P_kbar*1000)-1)/T_K
    fo2_NNO=10**logfo2_NNO



    logfo2_QFM=(-25096.3/T_K) + 8.735 + 0.11 * ((P_kbar*1000)-1)/T_K
    fo2_QFM=10**logfo2_QFM

    # This is Ln (XFe2O3/XFeO) from the Kress and Carmichael 1991 paper
    Z=np.log(mol_frac_hyd_redox['Fe2O3_Liq_mol_frac']/
         (mol_frac_hyd_redox['FeO_Liq_mol_frac']))

    # We've simplified the equatoin down to Z= a ln fo2 + rightside

    rightside=( (11492/T_K)-6.675+((-2.243*mol_frac_hyd_redox['Al2O3_Liq_mol_frac'])+(-1.828*hyd_mol_frac_test['FeOt_Liq_mol_frac'])
    +(3.201*mol_frac_hyd_redox['CaO_Liq_mol_frac'])+(5.854*mol_frac_hyd_redox['Na2O_Liq_mol_frac'])+(6.215*mol_frac_hyd_redox['K2O_Liq_mol_frac']))
    -3.36*(1-(To/T_K) - np.log(T_K/To)) -0.000000701*((P_kbar*100000000)/T_K)
    + -0.000000000154*(((T_K-1673)*(P_kbar*100000000))/T_K) + 0.0000000000000000385*((P_kbar*100000000)**2/T_K)
    )

    ln_fo2_calc=(Z-rightside)/0.196
    fo2_calc=np.exp(ln_fo2_calc)
    # and back to log base 10
    log_fo2_calc=np.log10(fo2_calc)
    DeltaQFM=log_fo2_calc-logfo2_QFM
    DeltaNNO=log_fo2_calc-logfo2_NNO


    liq_comps_c.insert(0, 'DeltaQFM', DeltaQFM)
    liq_comps_c.insert(1, 'DeltaNNO', DeltaNNO)
    liq_comps_c.insert(2, 'fo2_calc', fo2_calc)
    return liq_comps_c


import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy
import scipy.optimize as optimize
from scipy.special import erf

## Import data

df_ideal_liq = pd.DataFrame(columns=['SiO2_Liq', 'TiO2_Liq', 'Al2O3_Liq',
'FeOt_Liq', 'MnO_Liq', 'MgO_Liq', 'CaO_Liq', 'Na2O_Liq', 'K2O_Liq',
'P2O5_Liq', 'H2O_Liq', 'Fe3Fet_Liq', 'Ni_Liq_ppm', 'Cu_Liq_ppm'])

def import_data(filename, sheet_name=None, Petrolog=False, MELTS=False, sample_label=None, suffix=None):
    if 'csv' in filename:
        my_input = pd.read_csv(filename)

    if Petrolog is True:

        df=pd.read_excel(filename, sheet_name='Petrolog_Output_FRAC')
        df.columns= df.columns.str.replace(' ','',regex=True)
        df.columns= df.columns.str.replace('_melt','_Liq',regex=True)
        if sum(df.columns.str.contains('Ni_Liq'))>0:
            df['Ni_Liq_ppm']=df['Ni_Liq']
        else:
            df['Ni_Liq_ppm']=df['Ni_Liq']
            print('We didnt find a Ni column in your Petrolog input')

        if sum(df.columns.str.contains('Cu_Liq'))>0:
            df['Cu_Liq_ppm']=df['Cu_Liq']
        else:
            df['Cu_Liq_ppm']=df['Cu_Liq']
            print('We didnt find a Ni column in your Petrolog input')

        df['FeOt_Liq']=df['FeO_Liq']+df['Fe2O3_Liq']*0.89998
        df['Fe3Fet_Liq']=1-(df['FeO_Liq']/df['FeOt_Liq'])
        df['T_K']=df['Temperature']+273.15
        df['P_kbar']=df['Pressure(kbar)']
        my_input=df


    elif MELTS is True:
        df=pd.read_table(filename, sep=',')
        df.columns=df.columns.str.replace('wt% ','',regex=True)
        df2=df.rename(columns={
                                'SiO2': 'SiO2_Liq',
                                'TiO2': 'TiO2_Liq',
                                'CaO': 'CaO_Liq',
                                'Al2O3': 'Al2O3_Liq',
                                'FeO': 'FeO_Liq',
                                'Fe2O3': 'Fe2O3_Liq',
                                'Cr2O3': 'Cr2O3_Liq',
                                'MnO': 'MnO_Liq',
                                'MgO': 'MgO_Liq',
                                'NiO': 'NiO_Liq',
                                'Na2O': 'Na2O_Liq',
                                'K2O': 'K2O_Liq',
                                'P2O5': 'P2O5_Liq',
                                'H2O': 'H2O_Liq'
                                })

        df2['FeOt_Liq']=df2['FeO_Liq']+df2['Fe2O3_Liq']*0.89998
        df2['Fe3Fet_Liq']=1-(df2['FeO_Liq']/df2['FeOt_Liq'])
        df2['T_K']=df2['T (C)']+273.15
        df2['P_kbar']=df2['P (kbars)']
        my_input=df2

    elif 'xls' in filename:
        if sheet_name is not None:
            my_input = pd.read_excel(filename, sheet_name=sheet_name)
            #my_input[my_input < 0] = 0
        else:
            my_input = pd.read_excel(filename)
            #my_input[my_input < 0] = 0


    if suffix is not None:
        if any(my_input.columns.str.contains(suffix)):
            w.warn('We notice you have specified a suffix, but some of your columns already have this suffix. '
        'e.g., If you already have _Liq in the file, you shouldnt specify suffix="_Liq" during the import')

    my_input_c = my_input.copy()

    if "Sample_ID_Liq" not in my_input_c:
        my_input_c['Sample_ID_Liq'] = my_input.index

    if suffix is not None:
        my_input_c=my_input_c.add_suffix(suffix)

    if any(my_input.columns.str.contains("FeO_")) and (all(my_input.columns.str.contains("FeOt_")==False)):
        raise ValueError("No FeOt found. You've got a column heading with FeO. To avoid errors based on common EPMA outputs"
    " thermobar only recognises columns with FeOt for all phases except liquid"
    " where you can also enter a Fe3Fet_Liq heading used for equilibrium tests")

    if any(my_input.columns.str.contains("Fe2O3_")) and (all(my_input.columns.str.contains("FeOt_")==False)):
        raise ValueError("No FeOt column found. You've got a column heading with Fe2O3. To avoid errors based on common EPMA outputs"
        " thermobar only recognises columns with FeOt for all phases except liquid"
        " where you can also enter a Fe3Fet_Liq heading used for equilibrium tests")

    if any(my_input.columns.str.contains("FeOT_")) and (all(my_input.columns.str.contains("FeOt_")==False)):
        raise ValueError("No FeOt column found. You've got a column heading with FeOT. Change to a lower case t")



    myLiquids1 = my_input_c.reindex(df_ideal_liq.columns, axis=1).fillna(0)
    myLiquids1 = myLiquids1.apply(pd.to_numeric, errors='coerce').fillna(0)
    myLiquids1[myLiquids1 < 0] = 0
    print('We have replaced all missing liquid oxides and strings with zeros. ')

    cols2=myLiquids1.columns
    my_input_c=my_input.copy()
    for col in cols2:
        if col in my_input_c.columns:
            my_input_c=my_input_c.drop(columns=col)

    out=pd.concat([myLiquids1, my_input_c], axis=1)
    return out





import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy
import scipy.optimize as optimize
from scipy.special import erf



from PySulfSat.core_calcs import *
#SCAS (Li and Ripley, 2009;
#Baker and Moretti, 2011; Masotta and Keppler, 2015;
#Chowdhury and Dasgupta, 2019; Zajacz and Tsay, 2019)
df_ideal_liq_Chow = pd.DataFrame(columns=['SiO2_Liq', 'TiO2_Liq', 'Al2O3_Liq',
'FeOt_Liq', 'MnO_Liq', 'MgO_Liq', 'CaO_Liq', 'Na2O_Liq', 'K2O_Liq', 'H2O_Liq'])

def norm_liqs_with_H2O(Liqs):
    Liqs_c=Liqs.copy()
    Liqs_c=Liqs_c.fillna(0)
    Liqs2=Liqs_c.reindex(
        df_ideal_liq_Chow.columns, axis=1).fillna(0)

    sum_rows=Liqs2.sum(axis=1)

    Liqs_norm=Liqs2.divide(sum_rows/100, axis='rows')
    return Liqs_norm


def calculate_CD2019_SCAS(*, df, T_K, H2O_Liq=None):
    """" Calculates SCAS using the model of Chowdhury and Dasgupta, 2019


    Parameters
    -------
    df: pandas.DataFrame
        Dataframe of liquid compositions. Needs to have the headings "SiO2_Liq", "TiO2_Liq" etc, with
        compositions in oxide wt%. all FeO should be entered as FeOt_Liq

    T_K: int, float, pandas.Series
        Temperature in Kelvin.

    H2O_Liq: int, float, pandas.Series
        Option input, overwrites H2O_Liq in input dataframe




    Returns
    -------
    df of input dataframe, with new column heading "Calc SCAS (ppm)"

    """


    df_c=df.copy()
    df_norm=norm_liqs_with_H2O(Liqs=df_c)
    hyd_prop=calculate_hydrous_mol_proportions_liquid(liq_comps=df_norm)
    sum_hyd_prop=hyd_prop.sum(axis=1)
    hyd_frac=calculate_hydrous_mol_fractions_liquid(liq_comps=df_norm)
    sum_hydr_frac=hyd_frac.sum(axis=1)

    # Coefficients
    a=-13.23
    b=-0.5
    dsi=3.02
    dca=36.7
    dmg=2.84
    dfe=10.14
    dal=44.28
    dna=26.27
    dk=-25.77
    e=0.09
    f=0.54

    lnXS=(a+(b*(10**4/T_K))+dsi*hyd_frac['SiO2_Liq_mol_frac']
        +dca*hyd_frac['CaO_Liq_mol_frac']+dmg*hyd_frac['MgO_Liq_mol_frac']
        +dfe*hyd_frac['FeOt_Liq_mol_frac']+dal*hyd_frac['Al2O3_Liq_mol_frac']
        +dna*hyd_frac['Na2O_Liq_mol_frac']+dk*hyd_frac['K2O_Liq_mol_frac']
        +e*df_c['H2O_Liq']-f*np.log(hyd_frac['CaO_Liq_mol_frac']))
    Xs=np.exp(lnXS)
    molesS=Xs*(sum_hyd_prop+Xs)
    SCAS=molesS*32.065
    SCASppm=SCAS*10000
    df_c.insert(0, 'Calc SCAS (ppm)', SCASppm)
    return df_c

def calculate_ZT2022_SCAS(*, df, T_K, H2O_Liq=None):
    """" Calculates SCAS using the model of Zajacz and Tsay, 2022

    Parameters
    -------
    df: pandas.DataFrame
        Dataframe of liquid compositions. Needs to have the headings "SiO2_Liq", "TiO2_Liq" etc, with
        compositions in oxide wt%. all FeO should be entered as FeOt_Liq

    T_K: int, float, pandas.Series
        Temperature in Kelvin.

    H2O_Liq: int, float, pandas.Series
        Option input, overwrites H2O_Liq in input dataframe




    Returns
    -------
    df of input dataframe, with new column heading "Calc SCAS (ppm)"

    """


    df_c=df.copy()
    df_c['MnO_Liq']=0
    if H2O_Liq is not None:
        df_c['H2O_Liq']=H2O_Liq

    hyd_prop=calculate_hydrous_mol_proportions_liquid(liq_comps=df_c)
    sum_hyd_prop=hyd_prop.sum(axis=1)
    hyd_frac=calculate_hydrous_mol_fractions_liquid(liq_comps=df_c)
    sum_hydr_frac=hyd_frac.sum(axis=1)


    #NBO
    param1=(hyd_frac['Na2O_Liq_mol_frac']*2+hyd_frac['K2O_Liq_mol_frac']*2+2*(hyd_frac['CaO_Liq_mol_frac']
    +hyd_frac['MgO_Liq_mol_frac']+hyd_frac['FeOt_Liq_mol_frac'])-hyd_frac['Al2O3_Liq_mol_frac']*2)/(hyd_frac['Al2O3_Liq_mol_frac']*2+hyd_frac['SiO2_Liq_mol_frac'])
    param1_log=param1>0
    # If param1>0
    NBOT=(hyd_frac['Na2O_Liq_mol_frac']*2+hyd_frac['K2O_Liq_mol_frac']*2+2*(hyd_frac['CaO_Liq_mol_frac']+hyd_frac['MgO_Liq_mol_frac']+hyd_frac['FeOt_Liq_mol_frac'])-hyd_frac['Al2O3_Liq_mol_frac']*2)/(hyd_frac['Al2O3_Liq_mol_frac']*2+hyd_frac['SiO2_Liq_mol_frac'])
    NBOT[~param1_log]=0

    # P_Rhyolite
    param2=hyd_frac['K2O_Liq_mol_frac']+hyd_frac['Na2O_Liq_mol_frac']+hyd_frac['CaO_Liq_mol_frac']
    param2_log=param2>hyd_frac['Al2O3_Liq_mol_frac']
    P_Rhyo=3.11*(hyd_frac['K2O_Liq_mol_frac']+hyd_frac['Na2O_Liq_mol_frac']+hyd_frac['CaO_Liq_mol_frac']-hyd_frac['Al2O3_Liq_mol_frac'])
    P_Rhyo[~param2_log]=1.54*(hyd_frac['Al2O3_Liq_mol_frac']-(hyd_frac['CaO_Liq_mol_frac']+hyd_frac['Na2O_Liq_mol_frac']+hyd_frac['K2O_Liq_mol_frac']))

    # P_c
    P_c=(P_Rhyo+251*hyd_frac['CaO_Liq_mol_frac']**2+57*hyd_frac['MgO_Liq_mol_frac']**2+154*hyd_frac['FeOt_Liq_mol_frac']**2)/(2*hyd_frac['Al2O3_Liq_mol_frac']+hyd_frac['SiO2_Liq_mol_frac'])/(1+4.8*NBOT)
    P_T=np.exp(-7890/(T_K))
    P_H2O=hyd_frac['H2O_Liq_mol_frac']*(2.09-1.65*NBOT)+0.42*NBOT+0.23

    # Ksp
    Ksp=np.exp(1.226*np.log(P_c*P_T*P_H2O)+0.079)
    XSmelt=Ksp/hyd_frac['CaO_Liq_mol_frac']
    Smelt=XSmelt*sum_hyd_prop*32.07*10000

    df_c.insert(0, 'Calc SCAS (ppm)', Smelt)
    return df_c


__author__ = 'Penny Wieser'


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import warnings as w
import scipy.optimize as optimize
from scipy.special import erf


# This has the core calculations, e.g., molar fractions, cation fractions etc.
from PySulfSat.core_calcs import *
# This has the loading excel function
from PySulfSat.import_data import *
# This has the actual SCSS calculations
from PySulfSat.scss_calcs2 import *
# This has stuff for performing s6+ corrections
from PySulfSat.s6_corrections import *
from PySulfSat.scas_calc import *
from PySulfSat.sulf_mass_balance import *

#from PySCSS.SCSS import *
# version
from ._version import __version__





import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

def calculate_SCSS_Total(SCSS, S6St_Liq):
    '''
    Calculates SCSS Total from the SCSS (2-)
    See Wieser et al. (2019) for details

    Parameters
    -----------
    SCSS: int, float, pandas.Series
        SCSS from different calculators (only uses S2-)

    S6St_Liq: int, float, pandas.Series
        Proportion of S6+ in the liquid

    Returns
    -----------
    float, pandas.Series
        SCSS total, i.e inputted SCSS (2-) scaled up to account for S present in the S6+ form.

    '''
    SCSS_Total=SCSS/(1-S6St_Liq)
    return SCSS_Total

def calculate_S6St_Jugo2010(DeltaQFM):
    '''
    Calculates The S6/St ratio from DeltaQFM using
    Jugo et al. (2010)

    Parameters
    -----------
    DeltaQFM: int, float, pandas.Series
        Offset from the QFM buffer in log units.

    Returns
    -----------
    float, pandas.Series
        Proportion of S6+ in the liquid

    '''
    S6St_Liq=1/(1+10**(2.1-2*DeltaQFM))

    return S6St_Liq

def calculate_S6St_Nash2019(T_K, Fe3Fet_Liq):
    '''
    Calculates The S6/St ratio from the temperature and the proportion
    of Fe3+ in the liquid using Nash et al. (2019).

    Parameters
    -----------
    T_K: int, float, pandas.Series
        Temperature in Kelvin.

    Fe3Fet_Liq: int, float, pandas.Series
        Proportion of Fe3 in the liquid
    Returns
    -----------
    float, pandas.Series
        Proportion of S6+ in the liquid

    '''
    # First, calculate Fe3/Fe2 from Fe3/Fet
    Fe3Fe2=Fe3Fet_Liq/(1-Fe3Fet_Liq)
    log_Fe3Fe2=np.log10(Fe3Fe2)

    log_S6S2=8*log_Fe3Fe2 + 8.7436*10**6/T_K**2 - 27703/T_K + 20.273
    S6S2=10**(log_S6S2)

    S6=S6S2/(1+S6S2)

    S2=1/(1+S6S2)
    S6St=S6/(S6+S2)

    return S6St


def calculate_S_Tot_Kleinsasser2022_dacite(*, SCSS2=None,  SCAS=None, deltaFMQ=None):
    """ Calculates SCSS total as a function of S2- and S6+ in dacitic melts,
    following Kleinsasser et al. 2022"""

    if SCSS2 is not None:
        SCSS_Tot=SCSS2*(1+10**(2*deltaFMQ-3.05))
        return SCSS_Tot
    if SCAS is not None:
        SCAS_Tot=SCAS*(1+np.exp(1.26-2*deltaFMQ))
        return SCAS_Tot



def calculate_S_Tot_Jugo2010(*, SCSS2=None,  SCAS=None,DeltaQFM):
    '''Calculates S total as a function of S2- and S6+
    following Jugo et al. 2010 for given SCSS or SCAS

    Parameters
    -----------
    DeltaQFM: int, float, pandas.Series
        Offset from the QFM buffer in log units.

    Returns
    -----------
    float, pandas.Series
        Proportion of S6+ in the liquid

    '''
    if SCSS2 is not None:
        SCSS_Tot=SCSS2*(1+10**(2*deltaFMQ-2.1))
        return SCSS_Tot

        SCAS_Tot=SCAS*(1+np.exp(2.89-2.23*deltaFMQ))
        return SCAS_Tot







import matplotlib.pyplot as plt
import numpy as np
import pandas as pd


st_ratio=1/22.6436 #4.4163750000000000000E-02
 #1/22.220

def convert_d34_to_3432S(d34S=None):
    S3432=((d34S/1000)+1)*(st_ratio)
    return S3432

def convert_3432S_to_d34(S3432=None):
    d34S=(((S3432)/(st_ratio)) -1)*1000
    return d34S


def crystallize_S_incomp(S_init=1200, F_melt=None):
    """
    Calculates amount of S remaining in the melt for different melt fractoins (F_melt)
    assuming S is completely incompatible in silicate minerals


    Parameters
    --------------
    S_init: int, float, pd.Series
        initial S content in ppm
    F_melt: int, float, pd.Series
        melt fraction (between 0 and 1)


    Returns
    --------------
    S in the melt for each F

    """


    S_melt_FC=S_init/F_melt
    return S_melt_FC


def calculate_mass_frac_sulf(S_model=None, S_init=None, F_melt=None, S_sulf=None):
    """
    Calculates mass fraction of sulfide removed for a given SCSS value, a S content in the melt,
    and S content in the sulfide

    Parameters
    --------------
    S_model: int, float, pd.Series
        modelled amount of S present in the melt in ppm (could be SCSS2-, SCSStot, or SCAS, or STot from another method)
    S_init: int, float, pd.Series
        initial S content of the system where F_melt=1 in ppm
    F_melt:  int, float, pd.Series
        melt fraction (between 0 and 1)
    S_sulf:  int, float, pd.Series
        S content of the sulfide in ppm

    Returns
    --------------
    Mass fraction of sulfide

    """


    X_sulf=(S_init-F_melt*S_model)/S_sulf
    return X_sulf



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy
import scipy.optimize as optimize
from scipy.special import erf



from PySulfSat.core_calcs import *

## Li and Zhang 2022
df_ideal_liq_lizhang = pd.DataFrame(columns=['SiO2_Liq', 'TiO2_Liq', 'Al2O3_Liq',
'FeOt_Liq', 'MnO_Liq', 'MgO_Liq', 'CaO_Liq', 'Na2O_Liq', 'K2O_Liq',
'P2O5_Liq'])

def norm_liqs_excl_H2O(Liqs):
    Liqs_c=Liqs.copy()
    Liqs_c=Liqs_c.fillna(0)
    Liqs2=Liqs_c.reindex(
        df_ideal_liq_lizhang.columns, axis=1).fillna(0)

    sum_rows=Liqs2.sum(axis=1)

    Liqs_norm=Liqs2.divide(sum_rows/100, axis='rows')
    return Liqs_norm



def calculate_LZ2022_SCSS(*, df, T_K, P_kbar, H2O_Liq=None, Fe_FeNiCu_Sulf=None, Fe3Fet_Liq=None, logfo2=None,
Ni_Liq=None, Cu_Liq=None):

    '''
    Calculates SCSS using the model of Li and Zhang (2022), with options for users to
    calculate sulfide composition from liquid composition using the approach of Smythe et al. (2017),
    the empirical parameterization of O'Neill (2021), or input a sulfide composition.

    Parameters
    -------
    df: pandas.DataFrame
        Dataframe of liquid compositions. Needs to have the headings "SiO2_Liq", "TiO2_Liq" etc, with
        compositions in oxide wt%. all FeO should be entered as FeOt_Liq, which can then be partitioned
        using the Fe3Fet_Liq input. Heading order doesn't matter.

    T_K: int, float, pandas.Series
        Temperature in Kelvin.

    P_kbar: int, float, pandas.Series
        Pressure in kbar

    logfo2: int, float, or pandas.Series
        logfo2, used to convert to Fe3 following the method of Li and Zhang (2022)

    Fe3Fet_Liq: int, float, pandas.Series, or "Calc_ONeill"
        Proportion of Fe3+ in the liquid, as various parts of the calculation use only Fe2.
        If "Calc_ONeill" Calculates as a function of MgO content, using the MORB relationship.

    Sulfide Composition: Options to calculate from the liquid composition, enter the comp in el wt%,
    or enter the Fe_FeNiCu_Sulf, Cu

      if you want to calculate sulfide composition:

            Fe_FeNiCu_Sulf = "Calc_Smythe"
            Calculates sulfide composition analagous to the Solver function in the Smythe et al. (2017) spreadsheet.
            Here, we use Scipy optimize to find the ideal Ni and Cu
            contents using Kds from Kiseeva et al. (2015) and the Ni and Cu content of the melt.
            Requires user to also enter Ni_Liq and Cu_Liq.

            Or

            Fe_FeNiCu_Sulf = "Calc_ONeill"
            Calculates sulfide composition using the empirical expression of O'Neill (2021), which depends on
            FeOt_Liq, Ni_Liq, Cu_Liq, and Fe3Fet_Liq. We allow users to enter their own Fe3Fet_Liq,
            as we believe the empirical model of Neill where Fe3Fet_Liq is a function of MgO content is not
            broadly applicable.

        if you want to input a Fe_FeNiCu_Sulf ratio:
            Fe_FeNiCu_Sulf = int, float, pandas series
            Calculates SCSS using this ratio.
            If you want the non-ideal SCSS to be returned, you also need to enter
            values for Cu_FeNiCu_Sulf and Ni_FeNiCu_Sulf

        if you want to input a measured sulfide composition in el wt%
            Fe_Sulf, Ni_Sulf, Cu_Sulf = int, float, pandas series




    Returns
    -------
    pandas.DataFrame:
        Contains column for SCSS ideal, and user inputted dataframe

    '''



    T_K=T_K-0.15 # As they use 273, not 273.15 for their conversion

    Liqs=df.copy()

    if H2O_Liq is not None:
        Liqs['H2O_Liq']=H2O_Liq

    Liqs_norm=norm_liqs_excl_H2O(Liqs)

    moles=calculate_anhydrous_mol_proportions_liquid(liq_comps=Liqs_norm)
    mol_frac=calculate_anhydrous_mol_fractions_liquid(liq_comps=Liqs_norm)

    # Cation calculations

    if logfo2 is not None:
        logfo2=Liqs['logfO2']

        logXFe2O3XFeO=(0.196*logfo2/0.4343+11492/(T_K)-6.675-2.243*mol_frac['Al2O3_Liq_mol_frac']
                    -1.828*mol_frac['FeOt_Liq_mol_frac']+3.201*mol_frac['CaO_Liq_mol_frac']
                    +5.854*mol_frac['Na2O_Liq_mol_frac']+6.215*mol_frac['K2O_Liq_mol_frac']
                    -3.36*(1-1673/(T_K)-np.log((T_K)/1673))-0.0701*(P_kbar*1000)/(T_K)
                    -0.0000154*(P_kbar*1000)/(T_K)*(T_K-1673)+0.000000385*(P_kbar*1000)**2/(T_K))
        FeO_mol_frac=mol_frac['FeOt_Liq_mol_frac']/(1+2*np.exp(logXFe2O3XFeO))
        Fe2O3_mol_frac=mol_frac['FeOt_Liq_mol_frac']/(1+2*np.exp(logXFe2O3XFeO))*np.exp(logXFe2O3XFeO)

        mol_cat=pd.DataFrame(data={'Si_cat': moles['SiO2_Liq_mol_prop'],
                            'Ti_cat': moles['TiO2_Liq_mol_prop'],
                                'Al_cat': 2*moles['Al2O3_Liq_mol_prop'],
                            'Fe_cat':moles['FeOt_Liq_mol_prop']/(1+2*np.exp(logXFe2O3XFeO)),
                            'Mn_cat': moles['MnO_Liq_mol_prop'],
                            'Mg_cat': moles['MgO_Liq_mol_prop'],
                            'Ca_cat': moles['CaO_Liq_mol_prop'],
                            'Na_cat':2* moles['Na2O_Liq_mol_prop'],
                            'K_cat': 2* moles['K2O_Liq_mol_prop'],
                            'P_cat': 2* moles['P2O5_Liq_mol_prop'],
                            'H_cat': 0,
                            'Fe3_cat':moles['FeOt_Liq_mol_prop']/(1+2*np.exp(logXFe2O3XFeO))*2*np.exp(logXFe2O3XFeO) })

    if Fe3Fet_Liq is not None:
        mol_cat=pd.DataFrame(data={'Si_cat': moles['SiO2_Liq_mol_prop'],
                            'Ti_cat': moles['TiO2_Liq_mol_prop'],
                            'Al_cat': 2*moles['Al2O3_Liq_mol_prop'],
                            'Fe_cat':moles['FeOt_Liq_mol_prop']*(1-Fe3Fet_Liq),
                            'Mn_cat': moles['MnO_Liq_mol_prop'],
                            'Mg_cat': moles['MgO_Liq_mol_prop'],
                            'Ca_cat': moles['CaO_Liq_mol_prop'],
                            'Na_cat':2* moles['Na2O_Liq_mol_prop'],
                            'K_cat': 2* moles['K2O_Liq_mol_prop'],
                            'P_cat': 2* moles['P2O5_Liq_mol_prop'],
                            'H_cat': 0,
                            'Fe3_cat':moles['FeOt_Liq_mol_prop']*Fe3Fet_Liq})


    sum_mol_cat=mol_cat.sum(axis=1)
    mol_cat_norm=mol_cat.divide(sum_mol_cat, axis='rows')

    Fe3Fet_Liq=mol_cat_norm['Fe3_cat']/(mol_cat_norm['Fe_cat']+mol_cat_norm['Fe3_cat'])

    # print('Fe3_cat')
    # print(mol_cat_norm['Fe3_cat'])
    # print('Fe_cat')
    # print(mol_cat_norm['Fe_cat'])


    # Sulfide composition bit


    if isinstance(Fe_FeNiCu_Sulf, str):
        if Fe_FeNiCu_Sulf=="Calc_Smythe":


            # This does the Scipy minimisation of Cu and Ni contents using Kiseeva et al. (2015)
            calc_sulf=calculate_Symthe_sulf_minimisation(FeOt_Liq=Liqs['FeOt_Liq'], Fe3Fet_Liq=Fe3Fet_Liq,
                                            T_K=T_K, Ni_Liq=Ni_Liq, Cu_Liq=Cu_Liq)

            # This feeds those result back into a simpler function to get the Fe, S and O content of the sulfide
            Sulf_All=calculate_Kiseeva_sulf_comp_kd(Ni_Sulf=calc_sulf['Ni_Sulf'],Cu_Sulf=calc_sulf['Cu_Sulf'],
                                FeOt_Liq=df['FeOt_Liq'], Fe3Fet_Liq=Fe3Fet_Liq,
                                            T_K=T_K)

            Fe_FeNiCu_Sulf_calc=calculate_sulf_FeFeNiCu(Sulf_All['Ni_Sulf'], Sulf_All['Cu_Sulf'], Sulf_All['Fe_Sulf'])

        if Fe_FeNiCu_Sulf=="Calc_ONeill":
            Fe_FeNiCu_Sulf_calc=calculate_ONeill_sulf(FeOt_Liq=df['FeOt_Liq'],
            Ni_Liq=Ni_Liq, Cu_Liq=Cu_Liq, Fe3Fet_Liq=Fe3Fet_Liq)
    else:
        Fe_FeNiCu_Sulf_calc=Fe_FeNiCu_Sulf
        #print('replaced')







    NaKAl=mol_cat_norm['Na_cat']+mol_cat_norm['K_cat']-mol_cat_norm['Al_cat']
    DeltaGRT=(137778-91.666*(T_K)+8.474*(T_K)*np.log(T_K))/8.314/(T_K)

    SumXMAM=(1673/(T_K)*(6.7*(mol_cat_norm['Na_cat']+mol_cat_norm['K_cat'])
    +1.8*(mol_cat_norm['Al_cat']+mol_cat_norm['Fe3_cat'])+4.9*mol_cat_norm['Mg_cat']
    +8.1*mol_cat_norm['Ca_cat']+5*mol_cat_norm['Ti_cat']+8.9*(mol_cat_norm['Fe_cat']
    +mol_cat_norm['Mn_cat'])-22.2*(mol_cat_norm['Fe_cat']+
    mol_cat_norm['Mn_cat'])*mol_cat_norm['Ti_cat']+7.2*(mol_cat_norm['Fe_cat']
    +mol_cat_norm['Mn_cat'])*mol_cat_norm['Si_cat'])-2.06*erf(-7.2*(mol_cat_norm['Fe_cat']
    +mol_cat_norm['Mn_cat'])))

    lnCs=-23590/(T_K)+8.77+SumXMAM
    lnXFeO=-np.log(mol_cat_norm['Fe_cat'])
    LnrFeO=-((1-mol_cat_norm['Fe_cat'])**2*(28870-14710*mol_cat_norm['Mg_cat']
    +1960*mol_cat_norm['Ca_cat']+43300*mol_cat_norm['Na_cat']+95380*mol_cat_norm['K_cat']
    -76880*mol_cat_norm['Ti_cat'])+(1-mol_cat_norm['Fe_cat'])*(-62190*mol_cat_norm['Si_cat']
    +31520*mol_cat_norm['Si_cat']**2))/8.314/(T_K)
    # Need to sort out for low T<1200 C these two flip


    lnaFeS_lowT=-(31464-(T_K)*21.506)/8.314/(T_K)+np.log(Fe_FeNiCu_Sulf_calc)
    lnaFeS_HighT=np.log(1-mol_cat_norm['Fe_cat'])+np.log(Fe_FeNiCu_Sulf_calc)
    lnaFeS=lnaFeS_HighT
    lnaFeS[T_K<(1200+273)] =lnaFeS_lowT

    C1PC2erf_lowT=(-0.0291*(P_kbar)*1000+351*erf((P_kbar)*1000/10000))/(T_K)+0.04*(P_kbar)*1000/8.314/(T_K)
    C1PC2erf_highT=(-0.0291*(P_kbar)*1000+351*erf((P_kbar)*1000/10000))/(T_K)

    C1PC2erf=C1PC2erf_highT
    C1PC2erf[T_K<(1200+273)] =C1PC2erf_lowT

    lnS=DeltaGRT+lnCs+lnXFeO+LnrFeO+lnaFeS+C1PC2erf

    S2_calc=np.exp(lnS)
    S2_calc[(Liqs['FeOt_Liq']<5)&(Liqs['H2O_Liq']>0)]=0

    SumMoles_H2O=(moles['FeOt_Liq_mol_prop']+moles['MnO_Liq_mol_prop']+moles['MgO_Liq_mol_prop']
                +moles['CaO_Liq_mol_prop']+moles['Na2O_Liq_mol_prop']+moles['K2O_Liq_mol_prop'])
    XH2Ot=(Liqs['H2O_Liq']/18)/(Liqs['H2O_Liq']/18+(moles['SiO2_Liq_mol_prop']*2+moles['TiO2_Liq_mol_prop']*2+
    moles['Al2O3_Liq_mol_prop']*3+SumMoles_H2O+moles['P2O5_Liq_mol_prop']*5)*(100-Liqs['H2O_Liq'])/100)


    lnXH2Ot=np.log(XH2Ot)
    KOH=np.exp(2.6*mol_cat_norm['Si_cat']-4339*mol_cat_norm['Si_cat']/(T_K))
    XOH=(0.5-(0.25-(KOH-4)/KOH*(XH2Ot-XH2Ot**2))**0.5)/(KOH-4)*2*KOH
    lnXOH=np.log(XOH)
    XH2Om=XH2Ot-0.5*XOH
    lnXH2Om=np.log(XH2Om)
    lnXOH_XH2O=np.log(XOH+lnXH2Om)

    lnCHScalc=-19748*(1/T_K)+7.81+SumXMAM+lnXOH_XH2O

    HScal=(np.exp(DeltaGRT+lnCHScalc+lnXFeO+LnrFeO+lnaFeS+C1PC2erf))/(100+Liqs['H2O_Liq'])*100

    # If NaKAl>-0.015
    NaKAlterm=(1673/(T_K))*(19.634*NaKAl+0.2542)
    #Else
    LowNaKAl= NaKAl<-0.015
    NaKAlterm[LowNaKAl] =(1673/(T_K))*(26.365*NaKAl+0.9587)

    # If NaKAlterm>0
    lnCHS_NKA_term=lnCHScalc+NaKAlterm
    # Else
    lnCHS_NKA_term[NaKAlterm<=0]=lnCHScalc

    HScal2=(np.exp(DeltaGRT+lnCHS_NKA_term+lnXFeO+LnrFeO+lnaFeS+C1PC2erf))/(100+Liqs['H2O_Liq'])*100
    # If H2O>0

    # If HScal2>HScal
    Stot_calc_H2O=HScal2+S2_calc
    #If HScal2<=HScal
    calc2gcal=HScal2>HScal
    Stot_calc_H2O[~calc2gcal]=HScal+S2_calc

    # If H2O=0
    noH2O=~(Liqs['H2O_Liq']>0)
    Stot_calc_H2O[noH2O]=S2_calc

    Liqs.insert(0, 'SCSS_Tot', Stot_calc_H2O)
    Liqs.insert(1, 'Fe_FeNiCu_Sulf', Fe_FeNiCu_Sulf_calc)
    if any(Liqs.columns.str.contains('Fe3Fet_Liq')):
        if Fe3Fet_Liq is not None:
            print('replacing Fe3Fet_Liq in the original dataframe with that input into the function')
            Liqs['Fe3Fet_Liq']=Fe3Fet_Liq

    else:
        Liqs.insert(2, 'Fe3Fet_Liq', Fe3Fet_Liq)
    return Liqs



## Fortin et al. (2015) SCSS Calculation

def calculate_F2015_SCSS(*, df, T_K, P_kbar, H2O_Liq=None):
    '''
    Calculates SCSS using the model of Fortin et al. (2015).
    Unlike the models of Symthe et al. (2017) or O'Neill (2021) this model doesn't
    require the sulfide composition, or the proportion of Fe3 in the liqiud.

    Parameters
    -----------
    df: pandas.DataFrame
        Dataframe of liquid compositions from the import_data function.
        Needs to have the headings "SiO2_Liq", "TiO2_Liq" etc, with
        compositions in oxide wt%. all FeO should be entered as FeOt_Liq, which can then be partitioned
        using the Fe3Fet_Liq input. Heading order doesn't matter.

    T_K: int, float, pandas.Series
        Temperature in Kelvin.

    P_kbar: int, float, pandas.Series
        Pressure in kbar

    Returns
    -----------
    pandas.DataFrame
        Calculated SCSS, mol fractions, and input compositions.

    '''

    df_c=df.copy()
    if H2O_Liq is not None:
        df_c['H2O_Liq']=H2O_Liq
    mol_fracs=ss.calculate_hydrous_mol_fractions_liquid(df_c)
    Ln_SCSS_Calc=(34.7837+(1/T_K*-5772.322)+
((P_kbar/10)/T_K*-346.5377)+
(mol_fracs['SiO2_Liq_mol_frac']*-25.4986)+
(mol_fracs['TiO2_Liq_mol_frac']*-18.3435)+
(mol_fracs['Al2O3_Liq_mol_frac']*-27.3807)+
(mol_fracs['FeOt_Liq_mol_frac']*-17.2752)+
(mol_fracs['MnO_Liq_mol_frac']*0)+
(mol_fracs['MgO_Liq_mol_frac']*-22.3975)+
(mol_fracs['CaO_Liq_mol_frac']*-20.3778)+
(mol_fracs['Na2O_Liq_mol_frac']*-18.9539)+
(mol_fracs['K2O_Liq_mol_frac']*-32.1944)+
(mol_fracs['P2O5_Liq_mol_frac']*0)+
(mol_fracs['H2O_Liq_mol_frac']*-20.3934))
    SCSS_Calc=np.exp(Ln_SCSS_Calc)
    out=pd.concat([mol_fracs, df_c], axis=1)
    out.insert(0, 'SCSS_ppm_Fortin2015', SCSS_Calc)
    return out


## ONeill 2021 SCSS calculations

def calculate_ONeill_sulf(FeOt_Liq, Ni_Liq, Cu_Liq, MgO_Liq=None, Fe3Fet_Liq=None):
    '''
    Calculating the Fe_FeNiCu ratio in the sulfide using the empirical
    parameterizatin of ONeill.

    '''

    Fe_FeNiCu=1/(1+(Ni_Liq/(FeOt_Liq*(1-Fe3Fet_Liq)))*0.013+(Cu_Liq/(FeOt_Liq*(1-Fe3Fet_Liq)))*0.025)
    return Fe_FeNiCu


def calculate_O2021_SCSS(*, df, T_K, P_kbar, Fe3Fet_Liq=None, Fe_FeNiCu_Sulf=None,
Cu_FeNiCu_Sulf=None, Ni_FeNiCu_Sulf=None, Fe_Sulf=None, Ni_Sulf=None, Cu_Sulf=None,
Ni_Liq=None, Cu_Liq=None):

    '''
    Calculates SCSS using the model of O'Neill (2021), with options for users to
    calculate sulfide composition from liquid composition using the approach of Smythe et al. (2017),
    the empirical parameterization of O'Neill (2021),  or input a sulfide composition.

    Parameters
    -------
    df: pandas.DataFrame
        Dataframe of liquid compositions. Needs to have the headings "SiO2_Liq", "TiO2_Liq" etc, with
        compositions in oxide wt%. all FeO should be entered as FeOt_Liq, which can then be partitioned
        using the Fe3Fet_Liq input. Heading order doesn't matter.

    T_K: int, float, pandas.Series
        Temperature in Kelvin.

    P_kbar: int, float, pandas.Series
        Pressure in kbar

    Fe3Fet_Liq: int, float, pandas.Series, or "Calc_ONeill"
        Proportion of Fe3+ in the liquid, as various parts of the calculation use only Fe2.
        If "Calc_ONeill" Calculates as a function of MgO content, using the MORB relationship.

    Sulfide Composition: Options to calculate from the liquid composition, enter the comp in el wt%,
    or enter the FeFeNiCu, Cu

      if you want to calculate sulfide composition:

            Fe_FeNiCu_Sulf = "Calc_Smythe"
            Calculates sulfide composition analagous to the Solver function in the Smythe et al. (2017) spreadsheet.
            Here, we use Scipy optimize to find the ideal Ni and Cu
            contents using Kds from Kiseeva et al. (2015) and the Ni and Cu content of the melt.
            Requires user to also enter Ni_Liq and Cu_Liq.

            Or

            Fe_FeNiCu_Sulf = "Calc_ONeill"
            Calculates sulfide composition using the empirical expression of O'Neill (2021), which depends on
            FeOt_Liq, Ni_Liq, Cu_Liq, and Fe3Fet_Liq. We allow users to enter their own Fe3Fet_Liq,
            as we believe the empirical model of Neill where Fe3Fet_Liq is a function of MgO content is not
            broadly applicable.

        if you want to input a Fe_FeNiCu_Sulf ratio:
            Fe_FeNiCu_Sulf = int, float, pandas series
            Calculates SCSS using this ratio.
            If you want the non-ideal SCSS to be returned, you also need to enter
            values for Cu_FeNiCu_Sulf and Ni_FeNiCu_Sulf

        if you want to input a measured sulfide composition in el wt%
            Fe_Sulf, Ni_Sulf, Cu_Sulf = int, float, pandas series




    Returns
    -------
    pandas.DataFrame:
        Contains column for SCSS ideal, 1 sigma, input T and P, and the various intermediate steps of the calculation.

    '''
    df_c=df.copy()


    # if 'P2O5_Liq' in df_c.columns:
    df_c['P2O5_Liq']=0 # Doesnt have P2O5 in the input
    liqs=calculate_anhydrous_cat_fractions_liquid(liq_comps=df_c)

    if isinstance(Fe3Fet_Liq, str) and Fe3Fet_Liq == "Calc_ONeill":
        Fe2O3_Calc=np.exp(1.46-0.177*df_c['MgO_Liq'])
        Fe3Fet_Liq=Fe2O3_Calc*0.8998/(df_c['FeOt_Liq'])
        df_c['Fe3Fet_Liq']=Fe3Fet_Liq
        liqs['Fe3Fet_Liq_calc']=Fe3Fet_Liq
    if Fe3Fet_Liq is not None and not isinstance(Fe3Fet_Liq, str):
          df_c['Fe3Fet_Liq']=Fe3Fet_Liq



    liqs['Fe2_Liq_cat_frac']=liqs['Fet_Liq_cat_frac']*(1-Fe3Fet_Liq)

    # Then check options, basically need to have entered FeFeNiCu_Calc, or Fe_FeNiCu ratio, or all sulfide
    if Fe_Sulf is not None and Ni_Sulf is not None and Cu_Sulf is not None and Fe_FeNiCu_Sulf is not None:
        raise ValueError('You have entered both a Fe_FeNiCu_Sulf ratio, and the conc of Fe, Ni and Cu in your sulf. Please enter one set of inputs or another')

    elif isinstance(Fe_FeNiCu_Sulf, str) and (Fe_FeNiCu_Sulf=="Calc_Smythe" or Fe_FeNiCu_Sulf == "Calc_ONeill"):
        if Ni_Liq is None:
            raise ValueError('If you select Fe_FeNiCu_Sulf = Calc, you need to enter the concentration of Cu and Ni in the liquid in ppm"')
        if Cu_Liq is None:
            raise ValueError('If you select Fe_FeNiCu_Sulf = Calc, you need to enter the concentration of Cu and Ni in the liquid in ppm"')

    elif Fe_FeNiCu_Sulf is not None:
        print('Using inputted Fe_FeNiCu_Sulf ratio for calculations.')

    elif Fe_Sulf is not None and Ni_Sulf is not None and Cu_Sulf is not None:
        print('Using inputted Sulf compositions to calculate Fe_FeNiCu_Sulf ratios.')

    else:
        raise ValueError('Input for sulfide composition not recognised.')


    if isinstance(Fe_FeNiCu_Sulf, str):
        if Fe_FeNiCu_Sulf=="Calc_Smythe":


            # This does the Scipy minimisation of Cu and Ni contents using Kiseeva et al. (2015)
            calc_sulf=calculate_Symthe_sulf_minimisation(FeOt_Liq=df_c['FeOt_Liq'], Fe3Fet_Liq=df_c['Fe3Fet_Liq'],
                                            T_K=T_K, Ni_Liq=Ni_Liq, Cu_Liq=Cu_Liq)

            # This feeds those result back into a simpler function to get the Fe, S and O content of the sulfide
            Sulf_All=calculate_Kiseeva_sulf_comp_kd(Ni_Sulf=calc_sulf['Ni_Sulf'],Cu_Sulf=calc_sulf['Cu_Sulf'],
                                FeOt_Liq=df_c['FeOt_Liq'], Fe3Fet_Liq=df_c['Fe3Fet_Liq'],
                                            T_K=T_K)

            Fe_FeNiCu_Sulf_calc=calculate_sulf_FeFeNiCu(Sulf_All['Ni_Sulf'], Sulf_All['Cu_Sulf'], Sulf_All['Fe_Sulf'])
            Cu_FeNiCu_Sulf_calc=calculate_sulf_CuFeNiCu(Sulf_All['Ni_Sulf'], Sulf_All['Cu_Sulf'], Sulf_All['Fe_Sulf'])
            Ni_FeNiCu_Sulf_calc=calculate_sulf_NiFeNiCu(Sulf_All['Ni_Sulf'], Sulf_All['Cu_Sulf'], Sulf_All['Fe_Sulf'])

            liqs['Ni_Sulf_Calc']=Sulf_All['Ni_Sulf']
            liqs['Cu_Sulf_Calc']=Sulf_All['Cu_Sulf']
            liqs['Fe_Sulf_Calc']=Sulf_All['Fe_Sulf']
            liqs['O_Sulf_Calc']=Sulf_All['O_Sulf']
            liqs['S_Sulf_Calc']=Sulf_All['S_Sulf']

        if Fe_FeNiCu_Sulf=="Calc_ONeill":
            Fe_FeNiCu_Sulf_calc=calculate_ONeill_sulf(FeOt_Liq=df_c['FeOt_Liq'],
            Ni_Liq=Ni_Liq, Cu_Liq=Cu_Liq, Fe3Fet_Liq=Fe3Fet_Liq)
            liqs['Fe_FeNiCu_Sulf_calc']=Fe_FeNiCu_Sulf_calc
            Fe_FeNiCu_Sulf=Fe_FeNiCu_Sulf_calc

    elif Fe_Sulf is not None and Ni_Sulf is not None and Cu_Sulf is not None:
            Fe_FeNiCu_Sulf_calc=calculate_sulf_FeFeNiCu(Ni_Sulf, Cu_Sulf, Fe_Sulf)
            Cu_FeNiCu_Sulf_calc=calculate_sulf_CuFeNiCu(Ni_Sulf, Cu_Sulf, Fe_Sulf)
            Ni_FeNiCu_Sulf_calc=calculate_sulf_NiFeNiCu(Ni_Sulf, Cu_Sulf, Fe_Sulf)

    else:
        Fe_FeNiCu_Sulf_calc=Fe_FeNiCu_Sulf

    if isinstance(Fe_FeNiCu_Sulf_calc, float) is True:
        Fe_FeNiCu_Sulf_calc=Fe_FeNiCu_Sulf_calc
    else:
        Fe_FeNiCu_Sulf_calc=Fe_FeNiCu_Sulf_calc.astype(float)




    liqs['LnCS2_calc']=(8.77-23590/T_K+(1673/T_K)*(6.7*(liqs['Na_Liq_cat_frac']+liqs['K_Liq_cat_frac'])
    +4.9*liqs['Mg_Liq_cat_frac']+8.1*liqs['Ca_Liq_cat_frac']+8.9*(liqs['Fet_Liq_cat_frac']+liqs['Mn_Liq_cat_frac'])
    +5*liqs['Ti_Liq_cat_frac']+1.8*liqs['Al_Liq_cat_frac']
    -22.2*liqs['Ti_Liq_cat_frac']*(liqs['Fet_Liq_cat_frac']+liqs['Mn_Liq_cat_frac'])
        +7.2*(liqs['Fet_Liq_cat_frac']*liqs['Si_Liq_cat_frac']))-2.06*erf(-7.2*(liqs['Fet_Liq_cat_frac']+liqs['Mn_Liq_cat_frac'])))

    liqs['DeltaG']=((137778-91.666*T_K+8.474*T_K*np.log(T_K))/(8.31441*T_K)+(-291*(P_kbar/10)+351*erf((P_kbar/10)))/T_K)

    # print('Cation fractions')
    # print()


    liqs['Ln_a_FeS']=(np.log(Fe_FeNiCu_Sulf_calc*(1-liqs['Fe2_Liq_cat_frac'])))

    liqs['Ln_a_FeO']=( np.log(liqs['Fe2_Liq_cat_frac'])+(((1-liqs['Fe2_Liq_cat_frac'])**2)*
            (28870-14710*liqs['Mg_Liq_cat_frac']+1960*liqs['Ca_Liq_cat_frac']+43300*liqs['Na_Liq_cat_frac']+95380*liqs['K_Liq_cat_frac']-76880*liqs['Ti_Liq_cat_frac'])
            +(1-liqs['Fe2_Liq_cat_frac'])*(-62190*liqs['Si_Liq_cat_frac']+31520*liqs['Si_Liq_cat_frac']*liqs['Si_Liq_cat_frac']))/(8.31441*T_K))

    liqs['LnS']=(liqs['LnCS2_calc']+liqs['DeltaG']+liqs['Ln_a_FeS']-liqs['Ln_a_FeO'])

    liqs['SCSS2_ppm']=np.exp(liqs['LnS'])

    cols_to_move = ['SCSS2_ppm', 'LnS', "Ln_a_FeO",
                    'Ln_a_FeS', 'DeltaG', 'LnCS2_calc']

    Liqs = liqs[cols_to_move +
                                    [col for col in liqs.columns if col not in cols_to_move]]

    return Liqs




## Smythe SCSS calculations - More steps, as iteratively calculate sulfide composition.


def Loop_Smythe_sulf_calc_residual(single_argx0, FeO_Liq, T_K,  Ni_Liq, Cu_Liq):
    '''
    Calculates the residual between the Ni and Cu content of the Liquid and that of the sulfide for a
    Ni and Cu Kd calculated from Kiseeva et al. (2015).
    This function is used by the calculate_Smythe_sulf_minimisation function to calculate the sulfide compositoin,
    analagous to the Solver function in the supporting spreadsheet of Smythe et al.(2017).

    Parameters
    -----------
    single_argx0: float
        arguement for Ni and Cu in the sulfide to be solved by the scipy minimisation function function.

    FeO_Liq: int, float
        FeO (2+) content of the liquid in wt%

    Ni_Liq: int, float
        Ni content of the Liquid in ppm

    Cu_Liq: int, float
        Cu content of the Liquid in ppm

    T_K: int, float
        Temperature in Kelvin.


    Returns
    -----------
    float
        Calculated residual between measured Ni and Cu in the liquid, and that predicted
        from the sulfide composition and the Kd.
    '''
    #, x1, *, FeO_Liq, T_K,  Ni_Liq, Cu_Liq):
    Ni_Sulf=single_argx0[0]
    Cu_Sulf=single_argx0[1]

    OCalc_CellAG12=0.24*FeO_Liq
    FeS_calc_AL19=((100-Ni_Sulf-Cu_Sulf-OCalc_CellAG12-Cu_Sulf*20.1442646/79.8557354
    -Ni_Sulf*35.32650016/64.67349984))*36.47119049/100
    FeCalc_CellAF12=FeS_calc_AL19*63.52880951/36.47119049

    AG6=Ni_Sulf/58.6934/(Ni_Sulf/58.6934+Cu_Sulf/63.546+FeCalc_CellAF12/55.845)
    AG8=AG6*0.97
    AG9=AG6*0.92

        # If FeO<13
    if FeO_Liq<13:
        O_Sulf=(FeO_Liq*0.24*((1-AG8)**2))
    elif FeO_Liq>=13:
        O_Sulf=(FeO_Liq*0.24*((1-AG9)**2))

    AL17=(Ni_Sulf*35.32650016/64.67349984)
    AL18=(Cu_Sulf*20.1442646/79.8557354)
    AM20=O_Sulf*77.72983506/22.27016494

    S_Sulf=((100-AL17-AL18-(O_Sulf*77.72983506/22.27016494)
             -O_Sulf-Cu_Sulf-Ni_Sulf)*36.47119049/100 + AL17 + AL18)


    Fe_Sulf=((100-Ni_Sulf-Cu_Sulf-O_Sulf-AL17-AL18-AM20)
    *(63.52880951/100)+O_Sulf*77.72983506/22.27016494)

    NiSS=Ni_Sulf*(35.32650016/64.67349984)
    CuS05=Cu_Sulf*20.1442646/79.8557354
    OCalc=O_Sulf*77.72983506/22.27016494
    S_Sulf=(100-NiSS-CuS05-OCalc-O_Sulf-Cu_Sulf-Ni_Sulf)*36.47119049/100+NiSS+CuS05

    # Cu Coefficients
    A_Cu=0.6995
    B_Cu=4200
    eps_Cu_Cu=-1.0310
    eps_FeO_Cu=2.2030
    eps_Ni_Cu=0.00
    # Ni Coefficients

    A_Ni=1.869
    B_Ni=3300
    eps_Cu_Ni=0.00
    eps_Ni_Ni=-0.904
    eps_FeO_Ni=0




    Fe_Corr=FeO_Liq/(Fe_Sulf/55.845/(Ni_Sulf/58.6934+Cu_Sulf/63.546+Fe_Sulf/55.845)) #AF7
    Ni_FeNiCu=Ni_Sulf/58.6934/(Ni_Sulf/58.6934+Cu_Sulf/63.546+Fe_Sulf/55.845)
    Cu_FeNiCu=Cu_Sulf/63.546/(Ni_Sulf/58.6934+Cu_Sulf/63.546+Fe_Sulf/55.845)
    Fe_FeNiCu=Fe_Sulf/55.845/(Ni_Sulf/58.6934+Cu_Sulf/63.546+Fe_Sulf/55.845)

    DNi=(10**(B_Ni/T_K+A_Ni-np.log10(Fe_Corr)+(1673*eps_FeO_Ni*np.log10(1-0.049*O_Sulf)/T_K)
            +(1673*eps_Ni_Ni*np.log10(1-Ni_FeNiCu)/T_K)+(1673*eps_Cu_Ni*np.log10(1-Cu_FeNiCu)/T_K)))

    DCu=(10**(B_Cu/T_K+A_Cu-0.5*np.log10(Fe_Corr)+(1673*eps_FeO_Cu*np.log10(1-0.049*O_Sulf)/T_K)
             +(1673*eps_Cu_Cu*np.log10(1-Cu_FeNiCu)/T_K)+(1673*eps_Ni_Cu*np.log10(1-Ni_FeNiCu)/T_K)))

    Ni_Melt_calc=Ni_Sulf*10000/DNi
    Cu_Melt_calc=Cu_Sulf*10000/DCu

    Residual=(Ni_Melt_calc-Ni_Liq)**2 + (Cu_Melt_calc-Cu_Liq)**2
    return Residual


def calculate_Kiseeva_sulf_comp_kd(Ni_Sulf, Cu_Sulf, FeOt_Liq,  T_K, Fe3Fet_Liq=None):
    '''
    Calculates the Fe, O, S content of the sulfide using Kiseeva et al. (2015). Users can enter Ni, Cu and FeOt
    contents they have measured, or use this function once the
    Ni and Cu content of the sulfide have been calculated using scipy minimisation.
    Returns Kds, and S-O-Fe of sulfide.
    Also returns Se and Te calculated using Brenan, 2015

    Parameters
    -----------
    Ni_Sulf: pandas.Series
        Ni content of the sulfide in wt%, from the calculate_Smythe_sulf_minimisation function

    Cu_Sulf: pandas.Series
        Cu content of the sulfide in wt%, from the calculate_Smythe_sulf_minimisation function

    FeOt_Liq: pandas.Series
        FeOt content of the liquid in wt%, from the read_excel function.


    T_K: int, float, pandas.Series
        Temperature in Kelvin.

    P_kbar: int, float, pandas.Series
        Pressure in kbar

    Returns
    -----------
    pandas.DataFrame
        Calculated Cu, Ni, O, S and Fe content of sulfide.

    '''
    if Fe3Fet_Liq is not None:
        FeO_Liq=FeOt_Liq*(1-Fe3Fet_Liq)
    else:
        FeO_Liq=FeOt_Liq
    OCalc_CellAG12=0.24*FeO_Liq
    FeS_calc_AL19=((100-Ni_Sulf-Cu_Sulf-OCalc_CellAG12-Cu_Sulf*20.1442646/79.8557354
    -Ni_Sulf*35.32650016/64.67349984))*36.47119049/100
    FeCalc_CellAF12=FeS_calc_AL19*63.52880951/36.47119049

    AG6=Ni_Sulf/58.6934/(Ni_Sulf/58.6934+Cu_Sulf/63.546+FeCalc_CellAF12/55.845)
    AG8=AG6*0.97
    AG9=AG6*0.92

        # If FeO<13
    if isinstance(FeO_Liq, pd.Series):
        O_Sulf=(FeO_Liq*0.24*((1-AG8)**2))
        O_Sulf.loc[FeO_Liq>=13]=(FeO_Liq*0.24*((1-AG9)**2))
    if isinstance(FeO_Liq, float) or isinstance(FeO_Liq, int):
        if FeO_Liq>=13:
            O_Sulf=(FeO_Liq*0.24*((1-AG9)**2))
        else:
            O_Sulf=(FeO_Liq*0.24*((1-AG8)**2))


    AL17=(Ni_Sulf*35.32650016/64.67349984)
    AL18=(Cu_Sulf*20.1442646/79.8557354)
    AM20=O_Sulf*77.72983506/22.27016494

    S_Sulf=((100-AL17-AL18-(O_Sulf*77.72983506/22.27016494)
             -O_Sulf-Cu_Sulf-Ni_Sulf)*36.47119049/100 + AL17 + AL18)

    Fe_Sulf=((100-Ni_Sulf-Cu_Sulf-O_Sulf-AL17-AL18-AM20)
    *(63.52880951/100)+O_Sulf*77.72983506/22.27016494)

    NiSS=Ni_Sulf*(35.32650016/64.67349984)
    CuS05=Cu_Sulf*20.1442646/79.8557354
    OCalc=O_Sulf*77.72983506/22.27016494
    S_Sulf=(100-NiSS-CuS05-OCalc-O_Sulf-Cu_Sulf-Ni_Sulf)*36.47119049/100+NiSS+CuS05

    # Cu Coefficients
    eps_FeO_Cu=2.2030
    eps_Ni_Cu=0.00
    eps_Cu_Cu=-1.0310
    B_Cu=4200
    A_Cu=0.6995

    # Ni Coefficients
    eps_FeO_Ni=0
    eps_Ni_Ni=-0.904
    eps_Cu_Ni=0.00
    A_Ni=1.869
    B_Ni=3300

    # Pb Coefficients
    eps_FeO_Pb=0.454
    eps_Ni_Pb=0
    eps_Cu_Pb=0.00
    B_Pb=1260
    A_Pb=1.834


    # Ag Coefficients
    eps_FeO_Ag=1.901
    eps_Ni_Ag=0
    eps_Cu_Ag=0.00
    B_Ag=4300
    A_Ag=0.724

    # Zn Coefficients
    eps_FeO_Zn=-0.673
    eps_Ni_Zn=0.691
    eps_Cu_Zn=-1.023
    B_Zn=-990
    A_Zn=1.915

    # Cd Coefficients
    eps_FeO_Cd=0
    eps_Ni_Cd=0.48
    eps_Cu_Cd=-1.06
    B_Cd=1420
    A_Cd=1.919

    # Tl Coefficients
    eps_FeO_Tl=0.99
    eps_Ni_Tl=0.95
    eps_Cu_Tl=0.98
    B_Tl=0
    A_Tl=1.678

    # Mn Coefficients
    eps_FeO_Mn=-1.9
    eps_Ni_Mn=0.76
    eps_Cu_Mn=-0.41
    B_Mn=-1520
    A_Mn=1.629

    # In Coefficients
    eps_FeO_In=-1.61
    eps_Ni_In=-0.54
    eps_Cu_In=-2.36
    B_In=0
    A_In=2.502

    # Ti Coefficients
    eps_FeO_Ti=-12.91
    eps_Ni_Ti=-2.22
    eps_Cu_Ti=-2.04
    B_Ti=-2740
    A_Ti=1.027

    # Ga Coefficients
    eps_FeO_Ga=-5.09
    eps_Ni_Ga=0
    eps_Cu_Ga=-1.89
    B_Ga=-5470
    A_Ga=3.347


    # Sb Coefficients
    eps_FeO_Sb=-1.52
    eps_Ni_Sb=-2.32
    eps_Cu_Sb=-2.95
    B_Sb=-2670
    A_Sb=4.303

    # Co Coefficients
    eps_FeO_Co=0.60
    eps_Ni_Co=-0.28
    eps_Cu_Co=0.00
    B_Co=1280
    A_Co=1.964

    # V Cofficients
    eps_FeO_V=-5.34
    eps_Ni_V=0
    eps_Cu_V=0.00
    B_V=-2840
    A_V=2.524

    # Ge coefficients
    eps_FeO_Ge=-4.94
    eps_Ni_Ge=-1.65
    eps_Cu_Ge=-4.07
    B_Ge=-5000
    A_Ge=4.635

    # Cr coefficients
    eps_FeO_Cr=-0.76
    eps_Ni_Cr=0.44
    eps_Cu_Cr=0.00
    B_Cr=-1810
    A_Cr=2.356


    Fe_Corr=FeO_Liq/(Fe_Sulf/55.845/(Ni_Sulf/58.6934+Cu_Sulf/63.546+Fe_Sulf/55.845)) #AF7
    Ni_FeNiCu=Ni_Sulf/58.6934/(Ni_Sulf/58.6934+Cu_Sulf/63.546+Fe_Sulf/55.845)
    Cu_FeNiCu=Cu_Sulf/63.546/(Ni_Sulf/58.6934+Cu_Sulf/63.546+Fe_Sulf/55.845)
    Fe_FeNiCu=Fe_Sulf/55.845/(Ni_Sulf/58.6934+Cu_Sulf/63.546+Fe_Sulf/55.845)

    DNi=(10**(B_Ni/T_K+A_Ni-np.log10(Fe_Corr)+(1673*eps_FeO_Ni*np.log10(1-0.049*O_Sulf)/T_K)
            +(1673*eps_Ni_Ni*np.log10(1-Ni_FeNiCu)/T_K)+(1673*eps_Cu_Ni*np.log10(1-Cu_FeNiCu)/T_K)))

    DCu=(10**(B_Cu/T_K+A_Cu-0.5*np.log10(Fe_Corr)+(1673*eps_FeO_Cu*np.log10(1-0.049*O_Sulf)/T_K)
             +(1673*eps_Cu_Cu*np.log10(1-Cu_FeNiCu)/T_K)+(1673*eps_Ni_Cu*np.log10(1-Ni_FeNiCu)/T_K)))

    DPb=(10**(B_Pb/T_K+A_Pb-np.log10(Fe_Corr)
    +(1673*eps_FeO_Pb*np.log10(1-0.049*O_Sulf)/T_K)
    +(1673*eps_Ni_Pb*np.log10(1-Ni_FeNiCu)/T_K)
    +(1673*eps_Cu_Pb*np.log10(1-Cu_FeNiCu)/T_K)))

    DAg=(10**(B_Ag/T_K+A_Ag-0.5*np.log10(Fe_Corr)
    +(1673*eps_FeO_Ag*np.log10(1-0.049*O_Sulf)/T_K)
    +(1673*eps_Ni_Ag*np.log10(1-Ni_FeNiCu)/T_K)
    +(1673*eps_Cu_Ag*np.log10(1-Cu_FeNiCu)/T_K)))


    DZn=(10**(B_Zn/T_K+A_Zn-np.log10(Fe_Corr)
    +(1673*eps_FeO_Zn*np.log10(1-0.049*O_Sulf)/T_K)
     +(1673*eps_Ni_Zn*np.log10(1-Ni_FeNiCu)/T_K)
     +(1673*eps_Cu_Zn*np.log10(1-Cu_FeNiCu)/T_K)))

    DCd=(10**(B_Cd/T_K+A_Cd-np.log10(Fe_Corr)
    +(1673*eps_FeO_Cd*np.log10(1-0.049*O_Sulf)/T_K)
    +(1673*eps_Ni_Cd*np.log10(1-Ni_FeNiCu)/T_K)
    +(1673*eps_Cu_Cd*np.log10(1-Cu_FeNiCu)/T_K)))

    DTl=(10**(B_Tl/T_K+A_Tl-0.5*np.log10(Fe_Corr)
    +(1673*eps_FeO_Tl*np.log10(1-0.049*O_Sulf)/T_K)
     +(1673*eps_Ni_Tl*np.log10(1-Ni_FeNiCu)/T_K)
     +(1673*eps_Cu_Tl*np.log10(1-Cu_FeNiCu)/T_K)))

    DMn=(10**(B_Mn/T_K+A_Mn-np.log10(Fe_Corr)
    +(1673*eps_FeO_Mn*np.log10(1-0.049*O_Sulf)/T_K)
    +(1673*eps_Ni_Mn*np.log10(1-Ni_FeNiCu)/T_K)
    +(1673*eps_Cu_Mn*np.log10(1-Cu_FeNiCu)/T_K)))

    DIn=(10**(B_In/T_K+A_In-1.5*np.log10(Fe_Corr)
    +(1673*eps_FeO_In*np.log10(1-0.049*O_Sulf)/T_K)
     +(1673*eps_Ni_In*np.log10(1-Ni_FeNiCu)/T_K)
     +(1673*eps_Cu_In*np.log10(1-Cu_FeNiCu)/T_K)))

    DTi=(10**(B_Ti/T_K+A_Ti-2*np.log10(Fe_Corr)
    +(1673*eps_FeO_Ti*np.log10(1-0.049*O_Sulf)/T_K)
     +(1673*eps_Ni_Ti*np.log10(1-Ni_FeNiCu)/T_K)
     +(1673*eps_Cu_Ti*np.log10(1-Cu_FeNiCu)/T_K)))

    DGa=(10**(B_Ga/T_K+A_Ga-1.5*np.log10(Fe_Corr)
    +(1673*eps_FeO_Ga*np.log10(1-0.049*O_Sulf)/T_K)
     +(1673*eps_Ni_Ga*np.log10(1-Ni_FeNiCu)/T_K)
     +(1673*eps_Cu_Ga*np.log10(1-Cu_FeNiCu)/T_K)))

    DSb=(10**(B_Sb/T_K+A_Sb-1.5*np.log10(Fe_Corr)
    +(1673*eps_FeO_Sb*np.log10(1-0.049*O_Sulf)/T_K)
     +(1673*eps_Ni_Sb*np.log10(1-Ni_FeNiCu)/T_K)
     +(1673*eps_Cu_Sb*np.log10(1-Cu_FeNiCu)/T_K)))

    DCo=(10**(B_Co/T_K+A_Co-np.log10(Fe_Corr)
    +(1673*eps_FeO_Co*np.log10(1-0.049*O_Sulf)/T_K)
    +(1673*eps_Ni_Co*np.log10(1-Ni_FeNiCu)/T_K)
    +(1673*eps_Cu_Co*np.log10(1-Cu_FeNiCu)/T_K)))

    DV=(10**(B_V/T_K+A_V-1.5*np.log10(Fe_Corr)
    +(1673*eps_FeO_V*np.log10(1-0.049*O_Sulf)/T_K)
     +(1673*eps_Ni_V*np.log10(1-Ni_FeNiCu)/T_K)
     +(1673*eps_Cu_V*np.log10(1-Cu_FeNiCu)/T_K)))

    DGe=(10**(B_Ge/T_K+A_Ge-2*np.log10(Fe_Corr)
    +(1673*eps_FeO_Ge*np.log10(1-0.049*O_Sulf)/T_K)
     +(1673*eps_Ni_Ge*np.log10(1-Ni_FeNiCu)/T_K)
     +(1673*eps_Cu_Ge*np.log10(1-Cu_FeNiCu)/T_K)))

    DCr=(10**(B_Cr/T_K+A_Cr-np.log10(Fe_Corr)
    +(1673*eps_FeO_Cr*np.log10(1-0.049*O_Sulf)/T_K)
    +(1673*eps_Ni_Cr*np.log10(1-Ni_FeNiCu)/T_K)
    +(1673*eps_Cu_Cr*np.log10(1-Cu_FeNiCu)/T_K)))

    # Te and Se calculations using Brenan (2015) assuming only FeO
    DSe=(10**(3.47-3.07*10**(-2)*FeO_Liq - 9.13*10**(-4)*FeO_Liq**2
    ))
    DTe=(10**(4.3 -2.58*10**(-3)*FeO_Liq**2)
    )

    Ni_Melt_calc=Ni_Sulf*10000/DNi
    Cu_Melt_calc=Cu_Sulf*10000/DCu

    if isinstance(DNi, pd.Series):
        df_out=pd.DataFrame(data={
    'S_Sulf': S_Sulf, 'O_Sulf': O_Sulf, 'Fe_Sulf': Fe_Sulf, 'Ni_Sulf': Ni_Sulf, 'Cu_Sulf': Cu_Sulf, 'DNi': DNi,
    'DCu': DCu, 'DAg': DAg,  'DPb': DPb, 'DZn': DZn, 'DCd': DCd, 'DTl': DTl, 'DMn': DMn, 'DIn': DIn, 'DTi': DTi,
    'DGa': DGa, 'DSb': DSb, 'DCo': DCo, 'DV': DV, 'DGe': DGe, 'DCr': DCr, 'DSe_B2015': DSe,  'DTe_B2015': DTe})
    else:
        df_out=pd.DataFrame(data={
    'S_Sulf': S_Sulf, 'O_Sulf': O_Sulf, 'Fe_Sulf': Fe_Sulf, 'Ni_Sulf': Ni_Sulf, 'Cu_Sulf': Cu_Sulf, 'DNi': DNi,
    'DCu': DCu,  'DAg': DAg,  'DPb': DPb, 'DZn': DZn, 'DCd': DCd, 'DTl': DTl, 'DMn': DMn, 'DIn': DIn, 'DTi': DTi,
    'DGa': DGa, 'DSb': DSb, 'DCo': DCo, 'DV': DV, 'DGe': DGe , 'DCr': DCr, 'DSe_B2015': DSe,  'DTe_B2015': DTe}, index=[0])

    return df_out





## Calculating sulfide composition once you have iterated towards a respond.


def calculate_Symthe_sulf_minimisation(FeOt_Liq, Fe3Fet_Liq, T_K, Ni_Liq, Cu_Liq):

    '''
    Uses the Scipy optimize minimise function to find the Cu and Ni content of the sulfide
    that minimises the residual between liquid compositions, sulfide compositions and Kd,
    this minimis the Solver function in the supporting spreadsheet of Smythe et al.(2017).

    Parameters
    -----------
    FeOt_Liq: int, float, pd.Series
        FeOt content of the liquid in wt%

    Fe3Fet_Liq,: int, float, pd.Series
        Proportion of Fe3 in the liquid.

    Cu_Liq: int, float, pd.Series
        Cu content of the Liquid in ppm

    Ni_Liq: int, float, pd.Series
        Ni content of the Liquid in ppm

    T_K: int, float
        Temperature in Kelvin.


    Returns
    -----------
    float
        Calculated residual between measured Ni and Cu in the liquid, and that predicted
        from the sulfide composition and the Kd.
    '''

    Ni_Sulf=np.empty(len(FeOt_Liq), dtype=float)
    Cu_Sulf=np.empty(len(FeOt_Liq), dtype=float)
    bnds=((0, 30), (0, 30))

    for i in range(0, len(FeOt_Liq)):
        Calc_Sulf=scipy.optimize.minimize(Loop_Smythe_sulf_calc_residual, x0=(0, 0), bounds=bnds,
                        args=(FeOt_Liq.iloc[i]*(1-Fe3Fet_Liq.iloc[i]),
                              T_K.iloc[i], Ni_Liq.iloc[i], Cu_Liq.iloc[i])).get('x')



        Ni_Sulf[i]=Calc_Sulf[0]
        Cu_Sulf[i]=Calc_Sulf[1]
    if len(Ni_Sulf)>1:
        df_out=pd.DataFrame(data={'Ni_Sulf': Ni_Sulf, 'Cu_Sulf': Cu_Sulf})
    if len(Ni_Sulf)==1:
        df_out=pd.DataFrame(data={'Ni_Sulf': Ni_Sulf, 'Cu_Sulf': Cu_Sulf}, index=[0])

    return df_out


def calculate_sulf_FeFeNiCu(Ni_Sulf, Cu_Sulf, Fe_Sulf):
    '''
    Calculates the atomic Fe/(Fe+Ni+Cu) ratio of the sulfide.

    Parameters
    -----------
    Ni_Sulf: int, float, pd.Series
        Ni content of sulfide in wt%

    Cu_Sulf: int, float, pd.Series
        Cu content of sulfide in wt%

    Fe_Sulf: int, float, pd.Series
        Cu content of sulfide in wt%

    Returns
    -----------
    pd.Series, float
       Fe/(Fe+Ni+Cu) ratio of the sulfide


    '''

    Ni_moles=Ni_Sulf/58.6934
    Cu_moles=Cu_Sulf/63.546
    Fe_moles=Fe_Sulf/55.8457
    FeFeNiCu=Fe_moles/(Fe_moles+Cu_moles+Ni_moles)
    return FeFeNiCu

def calculate_sulf_CuFeNiCu(Ni_Sulf, Cu_Sulf, Fe_Sulf):
    '''
    Calculates the atomic Cu/(Fe+Ni+Cu) ratio of the sulfide.

    Parameters
    -----------
    Ni_Sulf: int, float, pd.Series
        Ni content of sulfide in wt%

    Cu_Sulf: int, float, pd.Series
        Cu content of sulfide in wt%

    Fe_Sulf: int, float, pd.Series
        Cu content of sulfide in wt%

    Returns
    -----------
    pd.Series, float
       Cu/(Fe+Ni+Cu) ratio of the sulfide


    '''
    Ni_moles=Ni_Sulf/58.6934
    Cu_moles=Cu_Sulf/63.546
    Fe_moles=Fe_Sulf/55.8457
    CuFeNiCu=Cu_moles/(Fe_moles+Cu_moles+Ni_moles)
    return CuFeNiCu

def calculate_sulf_NiFeNiCu(Ni_Sulf, Cu_Sulf, Fe_Sulf):
    '''
    Calculates the atomic Ni/(Fe+Ni+Cu) ratio of the sulfide.

    Parameters
    -----------
    Ni_Sulf: int, float, pd.Series
        Ni content of sulfide in wt%

    Cu_Sulf: int, float, pd.Series
        Cu content of sulfide in wt%

    Fe_Sulf: int, float, pd.Series
        Cu content of sulfide in wt%

    Returns
    -----------
    pd.Series, float
       Ni/(Fe+Ni+Cu) ratio of the sulfide


    '''
    Ni_moles=Ni_Sulf/58.6934
    Cu_moles=Cu_Sulf/63.546
    Fe_moles=Fe_Sulf/55.8457
    NiFeNiCu=Ni_moles/(Fe_moles+Cu_moles+Ni_moles)
    return NiFeNiCu
## Smythe Parameterization

def calculate_S2017_SCSS(*, df, T_K, P_kbar, Fe3Fet_Liq=None, Fe_FeNiCu_Sulf=None,
Cu_FeNiCu_Sulf=None, Ni_FeNiCu_Sulf=None, Fe_Sulf=None, Ni_Sulf=None, Cu_Sulf=None, Ni_Liq=None, Cu_Liq=None):
    '''
    Calculates SCSS using the model of Smythe et al. (2017), with options for users to
    calculate sulfide composition from liquid composition, or input sulfide composition.

    Parameters
    -------
    df: pandas.DataFrame
        Dataframe of liquid compositions. Needs to have the headings "SiO2_Liq", "TiO2_Liq" etc, with
        compositions in oxide wt%. all FeO should be entered as FeOt_Liq, which can then be partitioned
        using the Fe3Fet_Liq input. Heading order doesn't matter.

    T_K: int, float, pandas.Series
        Temperature in Kelvin.

    P_kbar: int, float, pandas.Series
        Pressure in kbar

    Fe3Fet_Liq: int, float, pandas.Series
        Proportion of Fe3+ in the liquid.
        Various parts of the calculation use only Fe2.

    Sulfide Composition: Options to calculate from the liquid composition, enter the comp in el wt%,
    or enter the FeFeNiCu, Cu

      if you want to calculate sulfide composition:

            Fe_FeNiCu_Sulf = "Calc_Smythe"
            Calculates sulfide composition analagous to the Solver function in the Smythe et al. (2017) spreadsheet.
            Here, we use Scipy optimize to find the ideal Ni and Cu
            contents using Kds from Kiseeva et al. (2015) and the Ni and Cu content of the melt.
            Requires user to also enter Ni_Liq and Cu_Liq.

            Or

            Fe_FeNiCu_Sulf = "Calc_ONeill"
            Calculates sulfide composition using the empirical expression of O'Neill (2021), which depends on
            FeOt_Liq, Ni_Liq, Cu_Liq, and Fe3Fet_Liq. We allow users to enter their own Fe3Fet_Liq,
            as we believe the empirical model of Neill where Fe3Fet_Liq is a function of MgO content is not
            broadly applicable.

        if you want to input a Fe_FeNiCu_Sulf ratio:
            Fe_FeNiCu_Sulf = int, float, pandas series
            Calculates SCSS using this ratio.
            If you want the non-ideal SCSS to be returned, you also need to enter
            values for Cu_FeNiCu_Sulf and Ni_FeNiCu_Sulf

        if you want to input a measured sulfide composition in el wt%
            Fe_Sulf, Ni_Sulf, Cu_Sulf = int, float, pandas series




    Returns
    -------
    pandas.DataFrame:
        Contains column for SCSS ideal, 1 sigma, input T and P, and the various intermediate steps of the calculation.

    '''
    df_c=df.copy()
    if Fe3Fet_Liq is not None:
        df_c['Fe3Fet_Liq']=Fe3Fet_Liq
    # First, calculate silicate hydrous mole fractions, as true regardless of choice of sulfide composition
    Smythe_calcs=calculate_Smythe_silicate_mole_fractions(df_c, Fe3Fet_Liq)

    # Then check options, basically need to have entered FeFeNiCu_Calc, or Fe_FeNiCu ratio, or all sulfide
    if Fe_Sulf is not None and Ni_Sulf is not None and Cu_Sulf is not None and Fe_FeNiCu_Sulf is not None:
        raise ValueError('You have entered both a Fe_FeNiCu_Sulf ratio, and the conc of Fe, Ni and Cu in your sulf. Please enter one set of inputs or another')

    elif isinstance(Fe_FeNiCu_Sulf, str) and (Fe_FeNiCu_Sulf=="Calc_Smythe" or Fe_FeNiCu_Sulf == "Calc_ONeill"):
        if Ni_Liq is None:
            raise ValueError('If you select Fe_FeNiCu_Sulf = Calc, you need to enter the concentration of Cu and Ni in the liquid in ppm')
        if Cu_Liq is None:
            raise ValueError('If you select Fe_FeNiCu_Sulf = Calc, you need to enter the concentration of Cu and Ni in the liquid in ppm')

    elif Cu_FeNiCu_Sulf is not None and Ni_FeNiCu_Sulf is not None and Fe_FeNiCu_Sulf is not None:
        Fe_FeNiCu_Sulf_calc=Fe_FeNiCu_Sulf
        Cu_FeNiCu_Sulf_calc=Cu_FeNiCu_Sulf
        Ni_FeNiCu_Sulf_calc=Ni_FeNiCu_Sulf

    elif Fe_FeNiCu_Sulf is not None and Cu_FeNiCu_Sulf is None and Ni_FeNiCu_Sulf is None:
        print('Using inputted Fe_FeNiCu_Sulf ratio for calculations.')
        Fe_FeNiCu_Sulf_calc=Fe_FeNiCu_Sulf
        Cu_FeNiCu_Sulf_calc =None

    elif Fe_Sulf is not None and Ni_Sulf is not None and Cu_Sulf is not None:
        print('Using inputted Sulf compositions to calculate Fe_FeNiCu_Sulf ratios.')

    else:
        raise ValueError('Input for sulfide composition not recognised.')

    if Fe_Sulf is not None and Ni_Sulf is not None and Cu_Sulf is not None:

        Fe_FeNiCu_Sulf_calc=calculate_sulf_FeFeNiCu(Ni_Sulf, Cu_Sulf, Fe_Sulf)
        Cu_FeNiCu_Sulf_calc=calculate_sulf_CuFeNiCu(Ni_Sulf, Cu_Sulf, Fe_Sulf)
        Ni_FeNiCu_Sulf_calc=calculate_sulf_NiFeNiCu(Ni_Sulf, Cu_Sulf, Fe_Sulf)



    elif isinstance(Fe_FeNiCu_Sulf, str):
        if Fe_FeNiCu_Sulf=="Calc_Smythe":


            # This does the Scipy minimisation of Cu and Ni contents using Kiseeva et al. (2015)
            calc_sulf=calculate_Symthe_sulf_minimisation(FeOt_Liq=df_c['FeOt_Liq'], Fe3Fet_Liq=df_c['Fe3Fet_Liq'],
                                            T_K=T_K, Ni_Liq=Ni_Liq, Cu_Liq=Cu_Liq)

            # This feeds those result back into a simpler function to get the Fe, S and O content of the sulfide
            Sulf_All=calculate_Kiseeva_sulf_comp_kd(Ni_Sulf=calc_sulf['Ni_Sulf'], Cu_Sulf=calc_sulf['Cu_Sulf'],
                                FeOt_Liq=df_c['FeOt_Liq'], Fe3Fet_Liq=df_c['Fe3Fet_Liq'],
                                            T_K=T_K)

            Fe_FeNiCu_Sulf_calc=calculate_sulf_FeFeNiCu(Sulf_All['Ni_Sulf'], Sulf_All['Cu_Sulf'], Sulf_All['Fe_Sulf'])
            Cu_FeNiCu_Sulf_calc=calculate_sulf_CuFeNiCu(Sulf_All['Ni_Sulf'], Sulf_All['Cu_Sulf'], Sulf_All['Fe_Sulf'])
            Ni_FeNiCu_Sulf_calc=calculate_sulf_NiFeNiCu(Sulf_All['Ni_Sulf'], Sulf_All['Cu_Sulf'], Sulf_All['Fe_Sulf'])

            Smythe_calcs['Ni_Sulf_Calc']=Sulf_All['Ni_Sulf']
            Smythe_calcs['Cu_Sulf_Calc']=Sulf_All['Cu_Sulf']
            Smythe_calcs['Fe_Sulf_Calc']=Sulf_All['Fe_Sulf']
            Smythe_calcs['O_Sulf_Calc']=Sulf_All['O_Sulf']
            Smythe_calcs['S_Sulf_Calc']=Sulf_All['S_Sulf']

        elif isinstance(Fe_FeNiCu_Sulf, str) and Fe_FeNiCu_Sulf=="Calc_ONeill":
            Fe_FeNiCu_Sulf_calc=calculate_ONeill_sulf(FeOt_Liq=df_c['FeOt_Liq'],
            Ni_Liq=Ni_Liq, Cu_Liq=Cu_Liq, Fe3Fet_Liq=Fe3Fet_Liq)
            df_c['Fe_FeNiCu_Sulf_calc']=Fe_FeNiCu_Sulf_calc
            Cu_FeNiCu_Sulf_calc=None
            Ni_FeNiCu_Sulf_calc=None



    # Calculating the different liquid components
    Smythe_calcs['Si_XA_ideal']=Smythe_calcs['Si_wt_atom']*(-27561.044)
    Smythe_calcs['Ti_XA_ideal']=Smythe_calcs['Ti_wt_atom']*(-11220.488)
    Smythe_calcs['Al_XA_ideal']=Smythe_calcs['Al_wt_atom']*(-18450.292)
    Smythe_calcs['Mg_XA_ideal']=Smythe_calcs['Mg_wt_atom']*(-13969.67)
    Smythe_calcs['Fe2_XA_ideal']=Smythe_calcs['Fe2_wt_atom']*(-34274.174)
    Smythe_calcs['Ca_XA_ideal']=Smythe_calcs['Ca_wt_atom']*(-7830.853)
    Smythe_calcs['Na_XA_ideal']=Smythe_calcs['Na_wt_atom']*(-13246.751)
    Smythe_calcs['K_XA_ideal']=Smythe_calcs['K_wt_atom']*(-29014.575)
    Smythe_calcs['H_XA_ideal']=Smythe_calcs['H_wt_atom']*(-17495.266)
    Smythe_calcs['Si*Fe_ideal']=(Smythe_calcs['Si_wt_atom']*Smythe_calcs['Fe2_wt_atom'])*116567.625


    Smythe_calcs['Si_XA_non_ideal']=Smythe_calcs['Si_wt_atom']*(-27996.431)
    Smythe_calcs['Ti_XA_non_ideal']=Smythe_calcs['Ti_wt_atom']*(-10714.991)
    Smythe_calcs['Al_XA_non_ideal']=Smythe_calcs['Al_wt_atom']*(-18999.945)
    Smythe_calcs['Mg_XA_non_ideal']=Smythe_calcs['Mg_wt_atom']*(-14512.488)
    Smythe_calcs['Fe2_XA_non_ideal']=Smythe_calcs['Fe2_wt_atom']*(-34895.294)
    Smythe_calcs['Ca_XA_non_ideal']=Smythe_calcs['Ca_wt_atom']*(-8831.616)
    Smythe_calcs['Na_XA_non_ideal']=Smythe_calcs['Na_wt_atom']*(-13712.715)
    Smythe_calcs['K_XA_non_ideal']=Smythe_calcs['K_wt_atom']*(-28583.983)
    Smythe_calcs['H_XA_non_ideal']=Smythe_calcs['H_wt_atom']*(-17766.114)
    Smythe_calcs['Si*Fe_non_ideal']=(Smythe_calcs['Si_wt_atom']*Smythe_calcs['Fe2_wt_atom'])*117815.515


    Smythe_calcs['log_SCSS_ideal']=(

(122175-80.28*T_K+8.474*T_K*np.log(T_K))/(8.314*T_K)+9.087+(Smythe_calcs['Si_XA_ideal']+Smythe_calcs['Ti_XA_ideal']
+Smythe_calcs['Al_XA_ideal']+Smythe_calcs['Mg_XA_ideal']+Smythe_calcs['Fe2_XA_ideal']+Smythe_calcs['Ca_XA_ideal']
+Smythe_calcs['Na_XA_ideal']+Smythe_calcs['K_XA_ideal']+Smythe_calcs['H_XA_ideal']+Smythe_calcs['Si*Fe_ideal'])/T_K
+np.log(Fe_FeNiCu_Sulf_calc)-np.log(Smythe_calcs['Fe2_wt_atom'])-269.4*0.1*P_kbar/T_K


    )


    Smythe_calcs['SCSS_ideal_ppm_Smythe2017']=np.exp(Smythe_calcs['log_SCSS_ideal'])
    Smythe_calcs['SCSS_ideal_ppm_Smythe2017_1sigma']=Smythe_calcs['SCSS_ideal_ppm_Smythe2017']*0.273169775211857
    Smythe_calcs['T_Input_K']=T_K
    Smythe_calcs['P_Input_kbar']=P_kbar
    Smythe_calcs['Fe_FeNiCu_Sulf']=Fe_FeNiCu_Sulf_calc
    Smythe_calcs['Fe3Fet_Liq_input']=Fe3Fet_Liq


    if Cu_FeNiCu_Sulf_calc is not None and Ni_FeNiCu_Sulf_calc is not None:
        Smythe_calcs['log_SCSS_non_ideal']=(
        (122175-80.28*T_K+8.474*T_K*np.log(T_K))/(8.314*T_K)+9.352+(Smythe_calcs['Si_XA_non_ideal']+Smythe_calcs['Ti_XA_non_ideal']
    +Smythe_calcs['Al_XA_non_ideal']+Smythe_calcs['Mg_XA_non_ideal']+Smythe_calcs['Fe2_XA_non_ideal']+Smythe_calcs['Ca_XA_non_ideal']
    +Smythe_calcs['Na_XA_non_ideal']+Smythe_calcs['K_XA_non_ideal']+Smythe_calcs['H_XA_non_ideal']+Smythe_calcs['Si*Fe_non_ideal'])
    /T_K+np.log(Fe_FeNiCu_Sulf_calc)-np.log(Smythe_calcs['Fe2_wt_atom'])-264.85*0.1*P_kbar/T_K+546.362*((Cu_FeNiCu_Sulf_calc**2 +Cu_FeNiCu_Sulf_calc*Ni_FeNiCu_Sulf_calc)/T_K)
        )
        Smythe_calcs['SCSS_non_ideal_ppm_Smythe2017']=np.exp(Smythe_calcs['log_SCSS_non_ideal'])
        Smythe_calcs['SCSS_non_ideal_ppm_Smythe2017_1sigma']=Smythe_calcs['SCSS_non_ideal_ppm_Smythe2017']*0.267299081373473

        cols_to_move = ['SCSS_ideal_ppm_Smythe2017', 'SCSS_ideal_ppm_Smythe2017_1sigma',
        'SCSS_non_ideal_ppm_Smythe2017', 'SCSS_non_ideal_ppm_Smythe2017_1sigma',
        'T_Input_K', "P_Input_kbar",'Fe_FeNiCu_Sulf', 'Fe3Fet_Liq_input']

    else:
        print('You havent entered a value for Ni_FeNiCu_Sulf and Cu_FeNiCu_Sulf so we cant calculate the non-ideal SCSS')
        cols_to_move = ['SCSS_ideal_ppm_Smythe2017', 'SCSS_ideal_ppm_Smythe2017_1sigma',
        'T_Input_K', "P_Input_kbar",'Fe_FeNiCu_Sulf', 'Fe3Fet_Liq_input']

    Smythe_calcs = Smythe_calcs[cols_to_move +
                                    [col for col in Smythe_calcs.columns if col not in cols_to_move]]

    Smythe_calcs['Fe_FeNiCu_Sulf_calc']=Fe_FeNiCu_Sulf_calc
    Concat=pd.concat([Smythe_calcs, df_c], axis=1)


    return Concat












